# pylint: disable=pointless-statement
from datetime import datetime, timedelta
from unittest.mock import patch
import pytest
from spark_expectations.core import get_spark_session
from spark_expectations.config.user_config import Constants as user_config
from spark_expectations.core.context import SparkExpectationsContext
from spark_expectations.core.exceptions import SparkExpectationsMiscException
from datetime import datetime, date
import os

spark = get_spark_session()


@patch("spark_expectations.core.context.uuid1")
def test_context_init(mock_uuid):
    # Test that the product_id is set correctly
    mock_uuid.return_value = "hghg-gjgu-jgj"
    context = SparkExpectationsContext(product_id="test_product", spark=spark)
    assert context.product_id == "test_product"

    # Test that the run_id is set correctly
    assert context.get_run_id == "test_product_hghg-gjgu-jgj"

    # Test that the run_date is set correctly
    run_date = context.get_run_date
    assert isinstance(datetime.strptime(run_date, "%Y-%m-%d %H:%M:%S"), datetime)


def test_context_properties():
    # Test that the getter properties return the correct values
    context = SparkExpectationsContext(product_id="product1", spark=spark)

    context._run_id = "test_run_id"
    context._run_date = "test_run_date"
    context._dq_stats_table_name = "test_dq_stats_table"
    context._dq_detailed_stats_table_name = "test_dq_stats_table"
    context._final_table_name = "test_final_table"
    context._error_table_name = "test_error_table"

    context._row_dq_rule_type_name = "row_dq_test"
    context._agg_dq_rule_type_name = "agg_dq_test"
    context._source_agg_dq_status = "test_skipped"
    context._row_dq_status = "test_skipped"
    context._final_agg_dq_status = "test_skipped"
    context._dq_run_status = "test_failed"
    context._se_enable_error_table = True
    context._dq_rules_params = {}

    context._dq_project_env_name = "APLAD-5063"
    context._dq_config_file_name = "dq_spark_expectations_config.ini"
    context._dq_config_abs_path = "sparkexpectations/config.ini"

    context._min_priority_email = "low"
    context._min_priority_pagerduty = "low"
    context._min_priority_slack = "low"
    context._min_priority_teams = "low"
    context._min_priority_zoom = "low"

    context._mail_smtp_server = "abc"
    context._mail_smtp_port = 25
    context._mail_smtp_password = "test_password"
    context._smtp_creds_dict = {
        "se.streaming.secret.type": "cerberus",
        "se.streaming.cerberus.url": "https://xyz.com",
        "se.streaming.cerberus.sdb.path": "abc",
        "spark.expectations.notifications.cerberus.smtp.password": "def",
    }
    context._enable_mail = True
    context._enable_smtp_server_auth = True
    context._enable_custom_email_body = True
    context._enable_templated_basic_email_body = True
    context._enable_templated_custom_email = True
    context._to_mail = "abc@mail.com, decf@mail.com"
    context._mail_from = "abc@mail.com"
    context._mail_subject = "spark expectations"
    context._email_custom_body = "test email body"
    context._enable_slack = True
    context._slack_webhook_url = "abcedfghi"
    context._enable_teams = True
    context._teams_webhook_url = "abcedfghi"
    context._enable_zoom = True
    context._zoom_webhook_url = "abcedfghi"
    context._zoom_token = "abcedfghi"
    context._table_name = "test_table"
    context._input_count = 100
    context._error_count = 10
    context._output_count = 90
    context._kafka_stats_topic_name = "spark_expectations_stats_topic"
    context._source_agg_dq_result = [
        {
            "action_if_failed": "ignore",
            "rule_type": "agg_dq",
            "rule_name": "sum_of_salary_threshold",
            "rule": "sum(salary)>100",
        }
    ]
    context._final_agg_dq_result = [
        {
            "action_if_failed": "ignore",
            "rule_type": "agg_dq",
            "rule_name": "sum_of_salary_threshold",
            "rule": "sum(salary)>100",
        }
    ]

    context._notification_on_start = True
    context._notification_on_completion = True
    context._enable_agg_dq_detailed_result = True

    context._enable_query_dq_detailed_result = True
    context._notification_on_fail = True
    context._env = "dev"
    context._error_drop_threshold = 100

    context._cerberus_url = "https://xyz"
    context._cerberus_cred_path = "spark-expectations/credentials"
    context._cerberus_token = "xxx"
    # context._kafka_bootstrap_server_url = "https://boostarp/server"
    # context._kafka_secret = "xxxx"
    # context._kafka_token_endpoint_uri = "https://token_uri"
    # context._kafka_client_id = "spark-expectations"

    context._run_id_name = "run_id"
    context._run_date_name = "run_date"
    context._run_date_time_name = "run_date_time"

    context._source_query_dq_status = "Passed"
    context._final_query_dq_status = "Skipped"

    context._debugger_mode = False

    context._supported_df_query_dq = spark.createDataFrame(
        [{"spark_expectations_query_check": "supported_place_holder_dataset_to_run_query_check"}]
    ).collect()

    datetime_now = datetime.now()

    context._source_agg_dq_start_time = datetime_now
    context._final_agg_dq_start_time = datetime_now
    context._source_query_dq_start_time = datetime_now
    context._final_query_dq_start_time = datetime_now
    context._row_dq_start_time = datetime_now
    context._dq_start_time = datetime_now

    context._source_agg_dq_end_time = datetime_now
    context._final_agg_dq_end_time = datetime_now
    context._source_query_dq_end_time = datetime_now
    context._final_query_dq_end_time = datetime_now
    context._row_dq_end_time = datetime_now
    context._dq_end_time = datetime_now

    context._num_row_dq_rules = 1
    context._num_dq_rules = 1
    context._num_agg_dq_rules = {"source_agg_dq": 1, "final_agg_dq": 1}
    context._num_query_dq_rules = {"source_query_dq": 1, "final_query_dq": 1}
    context._summarized_row_dq_res = [
        {"rule": "rule_1", "action_if_failed": "ignore", "failed_row_count": 2},
        {"rule": "rule_2", "action_if_failed": "fail", "failed_row_count": 4},
    ]

    context._kafka_row_dq_res_topic_name = "abc"

    context._se_streaming_stats_dict = {"a": "b", "c": "d"}
    context._se_streaming_stats_topic_name = "test_topic"
    context._stats_dict = [
        {
            "product_id": "test_product_id",
            "table_name": "table1",
        }
    ]

    assert context.get_run_id == "test_run_id"
    assert context.get_run_date == "test_run_date"
    assert context._dq_stats_table_name == "test_dq_stats_table"
    assert context._dq_detailed_stats_table_name == "test_dq_stats_table"
    assert context._final_table_name == "test_final_table"
    assert context._error_table_name == "test_error_table"

    assert context.get_row_dq_rule_type_name == "row_dq_test"
    assert context.get_agg_dq_rule_type_name == "agg_dq_test"
    assert context.get_source_agg_dq_status == "test_skipped"
    assert context.get_row_dq_status == "test_skipped"
    assert context.get_final_agg_dq_status == "test_skipped"
    assert context.get_dq_run_status == "test_failed"
    assert context.get_se_enable_error_table == True

    assert context._dq_project_env_name == "APLAD-5063"
    assert context._dq_config_file_name == "dq_spark_expectations_config.ini"
    assert context._dq_config_abs_path == "sparkexpectations/config.ini"
    assert context._mail_smtp_server == "abc"
    assert context.get_mail_smtp_port == 25
    assert context._mail_smtp_password == "test_password"
    assert context._smtp_creds_dict == {
        "se.streaming.secret.type": "cerberus",
        "se.streaming.cerberus.url": "https://xyz.com",
        "se.streaming.cerberus.sdb.path": "abc",
        "spark.expectations.notifications.cerberus.smtp.password": "def",
    }
    assert context._enable_mail is True
    assert context._enable_smtp_server_auth is True
    assert context._enable_custom_email_body is True
    assert context._enable_templated_basic_email_body is True
    assert context._enable_templated_custom_email is True
    assert context._to_mail == "abc@mail.com, decf@mail.com"
    assert context._mail_from == "abc@mail.com"
    assert context._mail_subject == "spark expectations"
    assert context._email_custom_body == "test email body"
    assert context._enable_slack is True
    assert context._slack_webhook_url == "abcedfghi"
    assert context._enable_teams is True
    assert context._teams_webhook_url == "abcedfghi"
    assert context._enable_zoom is True
    assert context._zoom_webhook_url == "abcedfghi"
    assert context._zoom_token == "abcedfghi"
    assert context._table_name == "test_table"
    assert context._input_count == 100
    assert context._error_count == 10
    assert context._output_count == 90
    assert context._kafka_stats_topic_name == "spark_expectations_stats_topic"
    assert context._source_agg_dq_result == [
        {
            "action_if_failed": "ignore",
            "rule_type": "agg_dq",
            "rule_name": "sum_of_salary_threshold",
            "rule": "sum(salary)>100",
        }
    ]
    assert context._final_agg_dq_result == [
        {
            "action_if_failed": "ignore",
            "rule_type": "agg_dq",
            "rule_name": "sum_of_salary_threshold",
            "rule": "sum(salary)>100",
        }
    ]

    assert context._notification_on_start is True
    assert context._notification_on_completion is True
    assert context._notification_on_fail is True
    assert context._env == "dev"
    assert context._error_drop_threshold == 100

    assert context._cerberus_url == "https://xyz"
    assert context.get_cerberus_url == "https://xyz"
    assert context._cerberus_cred_path == "spark-expectations/credentials"
    assert context.get_cerberus_cred_path == "spark-expectations/credentials"
    assert context._cerberus_token == "xxx"
    assert context.get_cerberus_token == "xxx"
    # assert context._kafka_bootstrap_server_url == "https://boostarp/server"
    # assert context._kafka_secret == "xxxx"
    # assert context._kafka_token_endpoint_uri == "https://token_uri"
    # assert context._kafka_client_id == "spark-expectations"

    assert context._run_id_name == "run_id"
    assert context._run_date_name == "run_date"
    assert context._run_date_time_name == "run_date_time"

    assert (
        context._supported_df_query_dq
        == spark.createDataFrame(
            [{"spark_expectations_query_check": "supported_place_holder_dataset_to_run_query_check"}]
        ).collect()
    )

    assert context._source_agg_dq_start_time == datetime_now
    assert context._final_agg_dq_start_time == datetime_now
    assert context._source_query_dq_start_time == datetime_now
    assert context._final_query_dq_start_time == datetime_now
    assert context._row_dq_start_time == datetime_now
    assert context._dq_start_time == datetime_now

    assert context._source_agg_dq_end_time == datetime_now
    assert context._final_agg_dq_end_time == datetime_now
    assert context._source_query_dq_end_time == datetime_now
    assert context._final_query_dq_end_time == datetime_now
    assert context._row_dq_end_time == datetime_now
    assert context._dq_end_time == datetime_now

    assert context._num_row_dq_rules == 1
    assert context._num_dq_rules == 1
    assert context._num_agg_dq_rules == {"source_agg_dq": 1, "final_agg_dq": 1}
    assert context._num_query_dq_rules == {"source_query_dq": 1, "final_query_dq": 1}
    assert context._summarized_row_dq_res == [
        {"rule": "rule_1", "action_if_failed": "ignore", "failed_row_count": 2},
        {"rule": "rule_2", "action_if_failed": "fail", "failed_row_count": 4},
    ]

    assert context._debugger_mode == False

    assert context._source_query_dq_status == "Passed"
    assert context._final_query_dq_status == "Skipped"
    assert context._kafka_row_dq_res_topic_name == "abc"
    assert context._se_streaming_stats_dict == {"a": "b", "c": "d"}
    assert context._se_streaming_stats_topic_name == "test_topic"
    assert context._stats_dict == [
        {
            "product_id": "test_product_id",
            "table_name": "table1",
        }
    ]


def test_set_dq_stats_table_name():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_dq_stats_table_name("dq_stats_table_name")
    assert context._dq_stats_table_name == "dq_stats_table_name"
    assert context.get_dq_stats_table_name == "dq_stats_table_name"


def test_set_final_table_name():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_final_table_name("final_table_name")
    assert context._final_table_name == "final_table_name"
    assert context.get_final_table_name == "final_table_name"


def test_error_table_name():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_error_table_name("error_table_name")
    assert context._error_table_name == "error_table_name"
    assert context.get_error_table_name == "error_table_name"


def test_row_dq_rule_type_name():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context._row_dq_rule_type_name = "row_dq1"
    context.get_row_dq_rule_type_name == "row_dq1"


def test_agg_dq_rule_type_name():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context._agg_dq_rule_type_name = "row_dq1"
    context.get_agg_dq_rule_type_name == "row_dq1"


def test_set_source_agg_dq_status():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_source_agg_dq_status("Passed")
    assert context._source_agg_dq_status == "Passed"
    assert context.get_source_agg_dq_status == "Passed"


def test_set_row_dq_status():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_row_dq_status("Failed")
    assert context._row_dq_status == "Failed"
    assert context.get_row_dq_status == "Failed"


def test_set_final_agg_dq_status():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_final_agg_dq_status("Skipped")
    assert context._final_agg_dq_status == "Skipped"
    assert context.get_final_agg_dq_status == "Skipped"


def test_set_dq_run_status():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_dq_run_status("Passed")
    assert context._dq_run_status == "Passed"
    assert context.get_dq_run_status == "Passed"


def test_get_source_agg_dq_status_exception():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context._source_agg_dq_status = None
    with pytest.raises(
        SparkExpectationsMiscException,
        match="The spark expectations context is not set completely, please assign "
        "'_source_agg_dq_status' before \n            accessing it",
    ):
        context.get_source_agg_dq_status


def test_get_row_dq_status_exception():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context._row_dq_status = None
    with pytest.raises(
        SparkExpectationsMiscException,
        match="The spark expectations context is not set completely, please assign "
        "'_row_dq_status' before \n            accessing it",
    ):
        context.get_row_dq_status


def test_get_final_agg_dq_status_exception():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context._final_agg_dq_status = None
    with pytest.raises(
        SparkExpectationsMiscException,
        match="The spark expectations context is not set completely, please assign "
        "'_final_agg_dq_status' before \n            accessing it",
    ):
        context.get_final_agg_dq_status


def test_get_dq_run_status_exception():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context._dq_run_status = None
    with pytest.raises(
        SparkExpectationsMiscException,
        match="The spark expectations context is not set completely, please assign "
        "'_dq_run_status' before \n            accessing it",
    ):
        context.get_dq_run_status


def test_get_row_dq_rule_type_name_exception():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context._row_dq_rule_type_name = None
    with pytest.raises(
        SparkExpectationsMiscException,
        match="The spark expectations context is not set completely, please assign "
        "'_row_dq_rule_type_name' before \n            accessing it",
    ):
        context.get_row_dq_rule_type_name


def test_get_agg_dq_rule_type_name_exception():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context._agg_dq_rule_type_name = None
    with pytest.raises(
        SparkExpectationsMiscException,
        match="The spark expectations context is not set completely, please assign "
        "'_agg_dq_rule_type_name' before \n            accessing it",
    ):
        context.get_agg_dq_rule_type_name


def test_set_source_query_dq_result():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    value = [{"1": "2"}]
    context.set_source_query_dq_result(value)
    context.get_source_query_dq_result == value


def test_set_final_query_dq_result():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    value = [{"1": "2"}]
    context.set_final_query_dq_result(value)
    context.get_final_query_dq_result == value


def test_get_query_dq_rule_type_name():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    values = [None, "query_dq"]
    for value in values:
        context._query_dq_rule_type_name = value
        if value is None:
            with pytest.raises(
                SparkExpectationsMiscException,
                match="The spark expectations context is not set completely, please assign "
                "'_query_dq_rule_type_name' before \n            accessing it",
            ):
                context.get_query_dq_rule_type_name
        else:
            context.get_query_dq_rule_type_name == value


def test_get_dq_stats_table_name_exception():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    with pytest.raises(
        SparkExpectationsMiscException,
        match="The spark expectations context is not set completely, please assign "
        "'_dq_stats_table_name' before \n            accessing it",
    ):
        context.get_dq_stats_table_name


def test_get_final_table_name_exception():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    with pytest.raises(
        SparkExpectationsMiscException,
        match="The spark expectations context is not set completely, please assign "
        "'_final_table_name' before \n            accessing it",
    ):
        context.get_final_table_name


def test_get_error_table_name_exception():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    with pytest.raises(
        SparkExpectationsMiscException,
        match="The spark expectations context is not set completely, please assign "
        "'_error_table_name' before \n            accessing it",
    ):
        context.get_error_table_name


def test_get_config_file_path():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context._dq_config_abs_path = "spark_expectations/config/file"

    assert context.get_config_file_path == "spark_expectations/config/file"


def test_get_config_file_path_exception():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context._dq_config_abs_path = None
    with pytest.raises(
        SparkExpectationsMiscException,
        match="""The spark expectations context is not set completely, please assign '_dq_config_abs_path' before
            accessing it""",
    ):
        context.get_config_file_path


def test_set_enable_mail():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_enable_mail(True)
    assert context._enable_mail is True
    assert context.get_enable_mail is True


def test_set_enable_smtp_server_auth():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_enable_smtp_server_auth(True)
    assert context._enable_smtp_server_auth is True
    assert context.get_enable_smtp_server_auth is True


def test_set_enable_custom_email_body():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_enable_custom_email_body(True)
    assert context._enable_custom_email_body is True
    assert context.get_enable_custom_email_body is True


def test_set_enable_templated_basic_email_body():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_enable_templated_basic_email_body(True)
    assert context._enable_templated_basic_email_body is True
    assert context.get_enable_templated_basic_email_body is True

def test_set_enable_templated_custom_email():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_enable_templated_custom_email(True)
    assert context._enable_templated_custom_email is True
    assert context.get_enable_templated_custom_email is True

def test_set_smtp_server():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_mail_smtp_server("abc")
    assert context._mail_smtp_server == "abc"
    assert context.get_mail_smtp_server == "abc"


def test_set_smtp_port():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_mail_smtp_port(25)
    context.set_mail_smtp_port(context._mail_smtp_port)
    assert context._mail_smtp_port == 25
    assert context.get_mail_smtp_port == 25


def test_set_mail_smtp_password():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_mail_smtp_password("test_password")
    assert context._mail_smtp_password == "test_password"
    assert context.get_mail_smtp_password == "test_password"


def test_set_smtp_creds_dict():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_smtp_creds_dict(
        {
            "se.streaming.secret.type": "cerberus",
            "se.streaming.cerberus.url": "https://xyz.com",
            "se.streaming.cerberus.sdb.path": "abc",
            "spark.expectations.notifications.cerberus.smtp.password": "def",
        }
    )
    assert context._smtp_creds_dict == {
        "se.streaming.secret.type": "cerberus",
        "se.streaming.cerberus.url": "https://xyz.com",
        "se.streaming.cerberus.sdb.path": "abc",
        "spark.expectations.notifications.cerberus.smtp.password": "def",
    }
    assert context.get_smtp_creds_dict == {
        "se.streaming.secret.type": "cerberus",
        "se.streaming.cerberus.url": "https://xyz.com",
        "se.streaming.cerberus.sdb.path": "abc",
        "spark.expectations.notifications.cerberus.smtp.password": "def",
    }


def test_set_to_mail():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_to_mail("abc@mail.com, def@mail.com")
    assert context._to_mail == "abc@mail.com, def@mail.com"
    assert context.get_to_mail == "abc@mail.com, def@mail.com"


def test_set_mail_from():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_mail_from("abc@mail.com")
    assert context._mail_from == "abc@mail.com"
    assert context.get_mail_from == "abc@mail.com"


def test_set_mail_subject():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_mail_subject("spark expectations")
    assert context._mail_subject == "spark expectations"
    assert context.get_mail_subject == "spark expectations"


def test_set_email_custom_body():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_email_custom_body("test email custom body")
    assert context._email_custom_body == "test email custom body"
    assert context.get_email_custom_body == "test email custom body"


def test_get_email_custom_body_exception():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context._email_custom_body = None
    with pytest.raises(
        SparkExpectationsMiscException,
        match="""The spark expectations context is not set completely, please assign '_email_custom_body' before 
            accessing it""",
    ):
        context.get_email_custom_body


def test_set_enable_slack():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_enable_slack(True)
    assert context._enable_slack is True
    assert context.get_enable_slack is True


def test_set_slack_webhook_url():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_slack_webhook_url("abcdefghi")
    assert context._slack_webhook_url == "abcdefghi"
    assert context.get_slack_webhook_url == "abcdefghi"


def test_set_enable_teams():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_enable_teams(True)
    assert context._enable_teams is True
    assert context.get_enable_teams is True


def test_set_teams_webhook_url():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_teams_webhook_url("abcdefghi")
    assert context._teams_webhook_url == "abcdefghi"
    assert context.get_teams_webhook_url == "abcdefghi"


def test_set_enable_zoom():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_enable_zoom(True)
    assert context._enable_zoom is True
    assert context.get_enable_zoom is True


def test_set_zoom_webhook_url():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_zoom_webhook_url("abcdefghi")
    assert context._zoom_webhook_url == "abcdefghi"
    assert context.get_zoom_webhook_url == "abcdefghi"


def test_set_zoom_token():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_zoom_token("abcdefghi")
    assert context._zoom_token == "abcdefghi"
    assert context.get_zoom_token == "abcdefghi"


def test_set_enable_pagerduty():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_enable_pagerduty(True)
    assert context._enable_pagerduty is True
    assert context.get_enable_pagerduty is True


def test_set_pagerduty_webhook_url():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_pagerduty_webhook_url("abcdefghi")
    assert context._pagerduty_webhook_url == "abcdefghi"
    assert context.get_pagerduty_webhook_url == "abcdefghi"


def test_set_pagerduty_integration_key():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_pagerduty_integration_key("abcdefghi")
    assert context._pagerduty_integration_key == "abcdefghi"
    assert context.get_pagerduty_integration_key == "abcdefghi"


def test_table_name():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_table_name("test_table")
    assert context._table_name == "test_table"
    assert context._table_name == "test_table"
    assert context.get_table_name == "test_table"


def test_set_input_count():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_input_count(100)
    assert context._input_count == 100
    assert context._input_count == 100
    assert context.get_input_count == 100


def test_set_error_count():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_error_count(10)
    assert context._error_count == 10
    assert context.get_error_count == 10


def test_set_output_count():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_output_count(90)
    assert context._output_count == 90
    assert context.get_output_count == 90


# def test_set_kafka_stats_topic_name():
#     context = SparkExpectationsContext(product_id="product1")
#     context.set_kafka_stats_topic_name("spark_expectations_stats_topic")
#     assert context._kafka_stats_topic_name == "spark_expectations_stats_topic"
#     assert context.get_kafka_stats_topic_name == "spark_expectations_stats_topic"


def test_set_kafka_source_agg_dq_result():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_source_agg_dq_result(
        [
            {
                "action_if_failed": "ignore",
                "rule_type": "agg_dq",
                "rule_name": "sum_of_salary_threshold",
                "rule": "sum(salary)>100",
            }
        ]
    )
    assert context._source_agg_dq_result == [
        {
            "action_if_failed": "ignore",
            "rule_type": "agg_dq",
            "rule_name": "sum_of_salary_threshold",
            "rule": "sum(salary)>100",
        }
    ]
    assert context.get_source_agg_dq_result == [
        {
            "action_if_failed": "ignore",
            "rule_type": "agg_dq",
            "rule_name": "sum_of_salary_threshold",
            "rule": "sum(salary)>100",
        }
    ]


def test_set_kafka_final_agg_dq_result():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_final_agg_dq_result(
        [
            {
                "action_if_failed": "ignore",
                "rule_type": "agg_dq",
                "rule_name": "sum_of_salary_threshold",
                "rule": "sum(salary)>100",
            }
        ]
    )
    assert context._final_agg_dq_result == [
        {
            "action_if_failed": "ignore",
            "rule_type": "agg_dq",
            "rule_name": "sum_of_salary_threshold",
            "rule": "sum(salary)>100",
        }
    ]
    assert context.get_final_agg_dq_result == [
        {
            "action_if_failed": "ignore",
            "rule_type": "agg_dq",
            "rule_name": "sum_of_salary_threshold",
            "rule": "sum(salary)>100",
        }
    ]


def test_get_mail_smtp_server_exception():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context._mail_smtp_server = None
    with pytest.raises(
        SparkExpectationsMiscException,
        match="The spark expectations context is not set completely, please assign "
        "'_mail_smtp_server' before \n            accessing it",
    ):
        context.get_mail_smtp_server


def test_get_mail_smtp_port_exception():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context._mail_smtp_port = 0
    with pytest.raises(
        SparkExpectationsMiscException,
        match="The spark expectations context is not set completely, please assign "
        "'_mail_smtp_port' before \n            accessing it",
    ):
        context.get_mail_smtp_port


def test_get_mail_smtp_password_none():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context._mail_smtp_password = None
    assert context.get_mail_smtp_password is None


def test_get_to_mail_exception():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context._to_mail = False
    with pytest.raises(
        SparkExpectationsMiscException,
        match="The spark expectations context is not set completely, please assign "
        "'_to_mail' before \n            accessing it",
    ):
        context.get_to_mail


def test_get_mail_subject_exception():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context._to_mail = False
    with pytest.raises(
        SparkExpectationsMiscException,
        match="The spark expectations context is not set completely, please assign "
        "'_mail_subject' before \n            accessing it",
    ):
        context.get_mail_subject


def test_get_mail_from_exception():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context._mail_from = False
    with pytest.raises(
        SparkExpectationsMiscException,
        match="The spark expectations context is not set completely, please assign "
        "'_mail_from' before \n            accessing it",
    ):
        context.get_mail_from


def test_get_slack_webhook_url_exception():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context._slack_webhook_url = False
    with pytest.raises(
        SparkExpectationsMiscException,
        match="The spark expectations context is not set completely, please assign "
        "'_slack_webhook_url' before \n            accessing it",
    ):
        context.get_slack_webhook_url


def test_get_teams_webhook_url_exception():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context._teams_webhook_url = False
    with pytest.raises(
        SparkExpectationsMiscException,
        match="The spark expectations context is not set completely, please assign "
        "'_teams_webhook_url' before \n            accessing it",
    ):
        context.get_teams_webhook_url


def test_get_zoom_webhook_url_exception():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context._zoom_webhook_url = False
    with pytest.raises(
        SparkExpectationsMiscException,
        match="The spark expectations context is not set completely, please assign "
        "'_zoom_webhook_url' before \n            accessing it",
    ):
        context.get_zoom_webhook_url


def test_get_pagerduty_webhook_url_exception():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context._pagerduty_webhook_url = False
    with pytest.raises(
        SparkExpectationsMiscException,
        match="The spark expectations context is not set completely, please assign "
        "'_pagerduty_webhook_url' before \n            accessing it",
    ):
        context.get_pagerduty_webhook_url


def test_get_zoom_token():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context._zoom_token = False
    with pytest.raises(
        SparkExpectationsMiscException,
        match="The spark expectations context is not set completely, please assign "
        "'_zoom_token' before \n            accessing it",
    ):
        context.get_zoom_token


def test_get_table_name_expection():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context._table_name = ""
    with pytest.raises(
        SparkExpectationsMiscException,
        match="The spark expectations context is not set completely, please assign "
        "'_table_name' before \n            accessing it",
    ):
        context.get_table_name


# def test_get_input_count():
#     context = SparkExpectationsContext(product_id="product1")
#     context._input_count = 0
#     with pytest.raises(SparkExpectationsMiscException,
#                        match="The spark expectations context is not set completely, please assign "
#                              "'_input_count' before \n            accessing it"):
#         context.get_input_count


# def test_get_kafka_stats_topic_name_exception():
#     context = SparkExpectationsContext(product_id="product1")
#     context._kafka_stats_topic_name = None
#     with pytest.raises(SparkExpectationsMiscException,
#                        match="The spark expectations context is not set completely, please assign "
#                              "'_kafka_stats_topic_name' before \n            accessing it"):
#         context.get_kafka_stats_topic_name


def test_set_notification_on_start():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_notification_on_start(True)
    assert context._notification_on_start is True
    assert context.get_notification_on_start is True


def test_set_notification_on_completion():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_notification_on_completion(True)
    assert context._notification_on_completion is True
    assert context.get_notification_on_completion is True


def test_set_notification_on_fail():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_notification_on_fail(True)
    assert context._notification_on_fail is True
    assert context.get_notification_on_fail is True


def test_set_env():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context._table_name = "dq_spark_staging.test_table"
    context.set_env("staging")
    assert context.get_env == "staging"


def test_get_env():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context._env = "dev1"
    assert context.get_env == "dev1"


def test_get_error_percentage():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context._input_count = 100
    context._error_count = 50

    assert context.get_error_percentage == 50.0


def test_get_output_percentage():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context._input_count = 100
    context._output_count = 50

    assert context.get_output_percentage == 50.0


def test_get_success_percentage():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context._input_count = 100
    context._output_count = 50
    context._error_count = 25

    assert context.get_success_percentage == 75.0


def test_get_error_drop_percentage():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context._input_count = 100
    context._output_count = 50
    context._error_count = 25

    assert context.get_error_drop_percentage == 50.0


def test_set_error_threshold():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_error_drop_threshold(100)
    assert context._error_drop_threshold == 100
    assert context.get_error_drop_threshold == 100


def test_get_error_threshold():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context._error_drop_threshold = 0
    with pytest.raises(
        SparkExpectationsMiscException,
        match="""The spark expectations context is not set completely, please assign '_error_drop_threshold'  before 
            accessing it""",
    ):
        context.get_error_drop_threshold


def test_get_cerberus_url_exception():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context._cerberus_url = None
    with pytest.raises(
        SparkExpectationsMiscException,
        match="""The spark expectations context is not set completely, please assign '_cerberus_url'  before 
            accessing it""",
    ):
        context.get_cerberus_url


def test_get_cerberus_cred_path_exception():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context._cerberus_cred_path = None
    with pytest.raises(
        SparkExpectationsMiscException,
        match="""The spark expectations context is not set completely, please assign '_cerberus_cred_path'  before 
            accessing it""",
    ):
        context.get_cerberus_cred_path


def test_get_cerberus_token_exception():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context._cerberus_token = None
    with pytest.raises(
        SparkExpectationsMiscException,
        match="""The spark expectations context is not set completely, please assign '_cerberus_token'  before 
            accessing it""",
    ):
        context.get_cerberus_token


# def test_get_kafka_bootstrap_server_url_exception():
#     context = SparkExpectationsContext(product_id="product1")
#     context._kafka_bootstrap_server_url = None
#     with pytest.raises(SparkExpectationsMiscException,
#                        match="""The spark expectations context is not set completely, please assign '_kafka_bootstrap_server_url'  before
#             accessing it"""):
#         context.get_kafka_bootstrap_server_url


# def test_get_kafka_secret_exception():
#     context = SparkExpectationsContext(product_id="product1")
#     context._kafka_secret = None
#     with pytest.raises(SparkExpectationsMiscException,
#                        match="""The spark expectations context is not set completely, please assign '_kafka_secret'  before
#             accessing it"""):
#         context.get_kafka_secret


# def test_get_kafka_token_endpoint_uri_exception():
#     context = SparkExpectationsContext(product_id="product1")
#     context._kafka_token_endpoint_uri = None
#     with pytest.raises(SparkExpectationsMiscException,
#                        match="""The spark expectations context is not set completely, please assign '_kafka_token_endpoint_uri'  before
#             accessing it"""):
#         context.get_kafka_token_endpoint_uri


# def test_get_kafka_client_id_exception():
#     context = SparkExpectationsContext(product_id="product1")
#     context._kafka_client_id = None
#     with pytest.raises(SparkExpectationsMiscException,
#                        match="""The spark expectations context is not set completely, please assign '_kafka_client_id'  before
#             accessing it"""):
#         context.get_kafka_client_id


# def test_set_kafka_bootstrap_server_url():
#     context = SparkExpectationsContext(product_id="product1")
#     context.set_kafka_bootstrap_server_url(kafka_bootstrap_server_url="https://boostarp/server")
#     assert context._kafka_bootstrap_server_url == "https://boostarp/server"
#     assert context.get_kafka_bootstrap_server_url == "https://boostarp/server"


# def test_kafka_secret():
#     context = SparkExpectationsContext(product_id="product1")
#     context.set_kafka_secret(kafka_secret="xxx")
#     assert context._kafka_secret == "xxx"
#     assert context.get_kafka_secret == "xxx"


# def test_kafka_token_endpoint_uri():
#     context = SparkExpectationsContext(product_id="product1")
#     context.set_kafka_token_endpoint_uri(kafka_token_endpoint_uri="https://token_uri")
#     assert context._kafka_token_endpoint_uri == "https://token_uri"
#     assert context.get_kafka_token_endpoint_uri == "https://token_uri"


# def test_set_kafka_client_id():
#     context = SparkExpectationsContext(product_id="product1")
#     context.set_kafka_client_id(kafka_client_id="spark-expectations")
#     assert context._kafka_client_id == "spark-expectations"
#     assert context.get_kafka_client_id == "spark-expectations"


def test_set_source_agg_dq_start_time():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_source_agg_dq_start_time()
    assert isinstance(context._source_agg_dq_start_time, datetime)


def test_set_source_agg_dq_end_time():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_source_agg_dq_end_time()
    assert isinstance(context._source_agg_dq_end_time, datetime)


def test_set_final_agg_dq_start_time():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_final_agg_dq_start_time()
    assert isinstance(context._final_agg_dq_start_time, datetime)


def test_set_final_agg_dq_end_time():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_final_agg_dq_end_time()
    assert isinstance(context._final_agg_dq_end_time, datetime)


def test_set_source_query_dq_start_time():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_source_query_dq_start_time()
    assert isinstance(context._source_query_dq_start_time, datetime)


def test_set_source_query_dq_end_time():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_source_query_dq_end_time()
    assert isinstance(context._source_query_dq_end_time, datetime)


def test_set_final_query_dq_start_time():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_final_query_dq_start_time()
    assert isinstance(context._final_query_dq_start_time, datetime)


def test_set_final_query_dq_end_time():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_final_query_dq_end_time()
    assert isinstance(context._final_query_dq_end_time, datetime)


def test_set_row_dq_start_time():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_row_dq_start_time()
    assert isinstance(context._row_dq_start_time, datetime)


def test_set_row_dq_end_time():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_row_dq_end_time()
    assert isinstance(context._row_dq_end_time, datetime)


def test_set_dq_start_time():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_dq_start_time()
    assert isinstance(context._dq_start_time, datetime)


def test_set_dq_end_time():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_dq_end_time()
    assert isinstance(context._dq_end_time, datetime)


def test_get_time_diff():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    assert context.get_time_diff(None, None) == 0.0
    now = datetime.now()
    assert context.get_time_diff(now, now + timedelta(seconds=2)) == 2.0


def test_get_source_agg_dq_run_time():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    now = datetime.now()
    context._source_agg_dq_start_time = now
    context._source_agg_dq_end_time = now + timedelta(seconds=2)
    assert context.get_source_agg_dq_run_time == 2.0


def test_get_final_agg_dq_run_time():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    now = datetime.now()
    context._final_agg_dq_start_time = now
    context._final_agg_dq_end_time = now + timedelta(seconds=2)
    assert context.get_final_agg_dq_run_time == 2.0


def test_get_source_query_dq_run_time():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    now = datetime.now()
    context._source_query_dq_start_time = now
    context._source_query_dq_end_time = now + timedelta(seconds=2)
    assert context.get_source_query_dq_run_time == 2.0


def test_get_final_query_dq_run_time():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    now = datetime.now()
    context._final_query_dq_start_time = now
    context._final_query_dq_end_time = now + timedelta(seconds=2)
    assert context.get_final_query_dq_run_time == 2.0


def test_get_row_dq_run_time():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    now = datetime.now()
    context._row_dq_start_time = now
    context._row_dq_end_time = now + timedelta(seconds=2)
    assert context.get_row_dq_run_time == 2.0


def test_get_dq_run_time():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    now = datetime.now()
    context._dq_start_time = now
    context._dq_end_time = now + timedelta(seconds=2)
    assert context.get_dq_run_time == 2.0


def test_get_dbr_version():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    os.environ["DATABRICKS_RUNTIME_VERSION"] = "13"
    assert context.get_dbr_version == 13

    os.environ["DATABRICKS_RUNTIME_VERSION"] = "13.3"
    assert context.get_dbr_version == 13.3

    # Remove the mock to test non-Databricks environment
    del os.environ["DATABRICKS_RUNTIME_VERSION"]
    assert context.get_dbr_version == None


def test_get_run_id_name():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    values = [None, "test"]
    for value in values:
        context._run_id_name = value
        if not value:
            with pytest.raises(
                SparkExpectationsMiscException,
                match="""The spark expectations context is not set completely, please assign '_run_id_name'  .*""",
            ):
                context.get_run_id_name
        else:
            context.get_run_id_name == value


def test_get_run_date_name():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    values = [None, "test"]
    for value in values:
        context._run_date_name = value
        if not value:
            with pytest.raises(
                SparkExpectationsMiscException,
                match="""The spark expectations context is not set completely, please assign '_run_date_name'  .*""",
            ):
                context.get_run_date_name
        else:
            context.get_run_date_name == value


def test_get_run_date_time_name():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    values = [None, "test"]
    for value in values:
        context._run_date_time_name = value
        if not value:
            with pytest.raises(
                SparkExpectationsMiscException,
                match="""The spark expectations context is not set completely, please assign '_run_date_time_name'  .*""",
            ):
                context.get_run_date_time_name
        else:
            context.get_run_date_time_name == value


def test_set_num_row_dq_rules():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_num_row_dq_rules()
    assert context._num_row_dq_rules == 1
    assert context._num_dq_rules == 1


def test_set_num_agg_dq_rules():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_num_agg_dq_rules(True, True)
    assert context.get_num_agg_dq_rules == {
        "num_source_agg_dq_rules": 1,
        "num_final_agg_dq_rules": 1,
        "num_agg_dq_rules": 1,
    }


def test_set_num_query_dq_rules():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_num_query_dq_rules(True, True)
    assert context.get_num_query_dq_rules == {
        "num_source_query_dq_rules": 1,
        "num_final_query_dq_rules": 1,
        "num_query_dq_rules": 1,
    }


def test_get_num_row_dq_rules():
    context = SparkExpectationsContext(product_id="product1", spark=spark)

    with pytest.raises(
        SparkExpectationsMiscException,
        match="""The spark expectations context is not set completely, please assign '_num_row_dq_rules'  .*""",
    ):
        context._num_row_dq_rules = None
        context.get_num_row_dq_rules

    context._num_row_dq_rules = 0
    context.set_num_row_dq_rules()
    context.get_num_row_dq_rules == 1


def test_get_num_agg_dq_rules_exception():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context._num_agg_dq_rules = [1, 2, 3, 4]
    with pytest.raises(
        SparkExpectationsMiscException,
        match="""The spark expectations context is not set completely, please assign '_num_agg_dq_rules'  before 
            accessing it""",
    ):
        context.get_num_agg_dq_rules


def test_get_num_query_dq_rules_exception():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context._num_query_dq_rules = [1, 2, 3, 4]
    with pytest.raises(
        SparkExpectationsMiscException,
        match="""The spark expectations context is not set completely, please assign '_num_query_dq_rules'  before 
            accessing it""",
    ):
        context.get_num_query_dq_rules


def test_get_num_dq_rules():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    with pytest.raises(
        SparkExpectationsMiscException,
        match="""The spark expectations context is not set completely, please assign '_num_dq_rules'  .*""",
    ):
        context._num_dq_rules = None
        context.get_num_dq_rules

    context.reset_num_dq_rules()
    context.get_num_dq_rules == 0


def test_set_summarized_row_dq_res():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_summarized_row_dq_res(
        [
            {"rule": "rule_1", "action_if_failed": "ignore", "failed_row_count": 2},
            {"rule": "rule_2", "action_if_failed": "fail", "failed_row_count": 4},
        ]
    )

    assert context.get_summarized_row_dq_res == [
        {"rule": "rule_1", "action_if_failed": "ignore", "failed_row_count": 2},
        {"rule": "rule_2", "action_if_failed": "fail", "failed_row_count": 4},
    ]


def test_set_target_and_error_table_writer_config():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_target_and_error_table_writer_config({"format": "bigquery"})

    assert context.get_target_and_error_table_writer_config == {"format": "bigquery"}


def test_set_stats_table_writer_config():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_stats_table_writer_config({"format": "bigquery"})

    assert context.get_stats_table_writer_config == {"format": "bigquery"}


# def test_set_kafka_row_dq_res_topic_name():
#     context = SparkExpectationsContext(product_id="product1")
#     context.set_kafka_row_dq_res_topic_name("abc")
#     assert context.get_kafka_row_dq_res_topic_name == "abc"
#
# def test_get_kafka_row_dq_res_topic_name_exception():
#     context = SparkExpectationsContext(product_id="product1")
#     context._kafka_row_dq_res_topic_name = None
#     with pytest.raises(SparkExpectationsMiscException,
#                        match= """The spark expectations context is not set completely, please assign '_kafka_row_dq_res_topic_name' before
#             accessing it"""):
#         context.get_kafka_row_dq_res_topic_name


def test_set_source_query_dq_status():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_source_query_dq_status("Passed")
    assert context.get_source_query_dq_status == "Passed"


def test_set_final_query_dq_status():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_final_query_dq_status("Passed")
    assert context.get_final_query_dq_status == "Passed"


def test_get_source_query_dq_status_exception():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context._source_query_dq_status = None
    with pytest.raises(
        SparkExpectationsMiscException,
        match="""The spark expectations context is not set completely, please assign '_source_query_dq_status' before 
            accessing it""",
    ):
        context.get_source_query_dq_status


def test_get_final_query_dq_status_exception():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context._final_query_dq_status = None
    with pytest.raises(
        SparkExpectationsMiscException,
        match="""The spark expectations context is not set completely, please assign '_final_query_dq_status' before 
            accessing it""",
    ):
        context.get_final_query_dq_status


def test_set_supported_df_query_dq():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context._supported_df_query_dq = context.set_supported_df_query_dq()
    assert (
        context.get_supported_df_query_dq.collect()
        == get_spark_session()
        .createDataFrame([{"spark_expectations_query_check": "supported_place_holder_dataset_to_run_query_check"}])
        .collect()
    )


def test_get_supported_df_query_dq():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context._supported_df_query_dq = None
    with pytest.raises(
        SparkExpectationsMiscException,
        match="""The spark expectations context is not set completely, please assign '_supported_df_query_dq'  before 
            accessing it""",
    ):
        context.get_supported_df_query_dq


def test_set_debugger_mode():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_debugger_mode(True)
    assert context._debugger_mode == True


def test_get_debugger_mode():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_debugger_mode(True)
    assert context.get_debugger_mode == True


def test_print_dataframe_with_debugger():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_debugger_mode(True)
    context.print_dataframe_with_debugger(context.set_supported_df_query_dq())


def test_get_error_percentage_negative():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context._input_count = 0
    assert context.get_error_percentage == 0.0


def test_get_error_percentage_negative():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context._input_count = 0
    assert context.get_error_percentage == 0.0


def test_get_output_percentage_negative():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context._input_count = 0
    assert context.get_output_percentage == 0.0


def test_get_success_percentage_negative():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context._input_count = 0
    assert context.get_success_percentage == 0.0


def test_get_error_drop_percentage_negative():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context._input_count = 0
    assert context.get_error_drop_percentage == 0.0


def test_reset_num_row_dq_rules():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.reset_num_row_dq_rules()
    assert context._num_row_dq_rules == 0


def test_reset_num_row_dq_rules():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.reset_num_agg_dq_rules()
    assert context._num_agg_dq_rules == {
        "num_agg_dq_rules": 0,
        "num_source_agg_dq_rules": 0,
        "num_final_agg_dq_rules": 0,
    }


def test_reset_num_query_dq_rules():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.reset_num_query_dq_rules()
    assert context._num_query_dq_rules == {
        "num_query_dq_rules": 0,
        "num_source_query_dq_rules": 0,
        "num_final_query_dq_rules": 0,
    }


def test_set_end_time_when_dq_job_fails():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    attributes = ["source_agg", "source_query", "row", "final_agg", "final_query"]
    for attribute in attributes:
        setattr(context, f"_{attribute}_dq_start_time", datetime.now())
        setattr(context, f"_{attribute}_dq_end_time", None)
        context.set_end_time_when_dq_job_fails()
        datetime_actual = getattr(context, f"_{attribute}_dq_end_time")
        datetime_actual.date == date.today()


def test_reset_num_dq_rules():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.reset_num_dq_rules()
    assert context._num_dq_rules == 0


def test_set_se_streaming_stats_dict():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_se_streaming_stats_dict({"a": "b", "c": "d"})

    assert context.get_se_streaming_stats_dict == {"a": "b", "c": "d"}


def get_set_se_streaming_stats_dict():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_se_streaming_stats_dict({"a": "b", "c": "d"})

    assert context.get_se_streaming_stats_dict == context._se_streaming_stats_dict


def test_set_stats_dict():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    data = [("test_product_id", "test_schema.table1", 5)]
    columns = ["product_id", "table_name", "error_count"]
    df = spark.createDataFrame(data, columns)
    context.set_stats_dict(df)

    assert context._stats_dict == [
        {
            "product_id": "test_product_id",
            "table_name": "test_schema.table1",
            "error_count": 5,
        }
    ]

    assert context.get_stats_dict == [
        {
            "product_id": "test_product_id",
            "table_name": "test_schema.table1",
            "error_count": 5,
        }
    ]


def test_get_secret_type():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_se_streaming_stats_dict({user_config.secret_type: "a"})

    assert context.get_secret_type == "a"


def test_get_secret_type_exception():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_se_streaming_stats_dict({user_config.se_enable_streaming: "a"})

    with pytest.raises(
        SparkExpectationsMiscException,
        match="""The spark expectations context is not set completely, please assign 
            'UserConfig.secret_type' before 
            accessing it""",
    ):
        context.get_secret_type


def test_get_server_url_key():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_se_streaming_stats_dict({user_config.dbx_kafka_server_url: "b", user_config.secret_type: "databricks"})

    assert context.get_server_url_key == "b"

    context.set_se_streaming_stats_dict(
        {
            user_config.dbx_kafka_server_url: "b",
            user_config.cbs_kafka_server_url: "c",
            user_config.secret_type: "cerberus",
        }
    )
    assert context.get_server_url_key == "c"


def test_get_server_url_key_exception():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_se_streaming_stats_dict({user_config.dbx_kafka_server_url: "b", user_config.secret_type: "cerberus"})

    with pytest.raises(
        SparkExpectationsMiscException,
        match="""The spark expectations context is not set completely, please assign
            'UserConfig.cbs_kafka_server_url' before
            accessing it""",
    ):
        context.get_server_url_key


def test_get_token_endpoint_url():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_se_streaming_stats_dict({user_config.dbx_secret_token_url: "d", user_config.secret_type: "databricks"})

    assert context.get_token_endpoint_url == "d"

    context.set_se_streaming_stats_dict(
        {
            user_config.dbx_secret_token_url: "d",
            user_config.cbs_secret_token_url: "f",
            user_config.secret_type: "cerberus",
        }
    )
    assert context.get_token_endpoint_url == "f"


def test_get_token_endpoint_url_exception():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_se_streaming_stats_dict({user_config.dbx_secret_token_url: "d", user_config.secret_type: "cerberus"})

    with pytest.raises(
        SparkExpectationsMiscException,
        match="""The spark expectations context is not set completely, please assign
            'UserConfig.cbs_secret_token_url' before
            accessing it""",
    ):
        context.get_token_endpoint_url


def test_get_token():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_se_streaming_stats_dict({user_config.dbx_secret_token: "g", user_config.secret_type: "databricks"})

    assert context.get_token == "g"

    context.set_se_streaming_stats_dict(
        {
            user_config.dbx_secret_token: "g",
            user_config.cbs_secret_token: "h",
            user_config.secret_type: "cerberus",
        }
    )
    assert context.get_token == "h"


def test_get_token_exception():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_se_streaming_stats_dict({user_config.dbx_secret_token_url: "g", user_config.secret_type: "cerberus"})

    with pytest.raises(
        SparkExpectationsMiscException,
        match="""The spark expectations context is not set completely, please assign
            'UserConfig.cbs_secret_token' before
            accessing it""",
    ):
        context.get_token


def test_get_client_id():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_se_streaming_stats_dict({user_config.dbx_secret_app_name: "i", user_config.secret_type: "databricks"})

    assert context.get_client_id == "i"

    context.set_se_streaming_stats_dict(
        {
            user_config.dbx_secret_app_name: "i",
            user_config.cbs_secret_app_name: "j",
            user_config.secret_type: "cerberus",
        }
    )
    assert context.get_client_id == "j"


def test_get_client_id_exception():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_se_streaming_stats_dict({user_config.dbx_secret_app_name: "g", user_config.secret_type: "cerberus"})

    with pytest.raises(
        SparkExpectationsMiscException,
        match="""The spark expectations context is not set completely, please assign
            'UserConfig.cbs_secret_app_name' before
            accessing it""",
    ):
        context.get_client_id


def test_get_topic_name():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_se_streaming_stats_dict({user_config.dbx_topic_name: "k", user_config.secret_type: "databricks"})

    assert context.get_topic_name == "k"

    context.set_se_streaming_stats_dict(
        {
            user_config.dbx_topic_name: "k",
            user_config.cbs_topic_name: "l",
            user_config.secret_type: "cerberus",
        }
    )
    assert context.get_topic_name == "l"


def test_get_topic_name_exception():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_se_streaming_stats_dict({user_config.dbx_topic_name: "k", user_config.secret_type: "cerberus"})

    with pytest.raises(
        SparkExpectationsMiscException,
        match="""The spark expectations context is not set completely, please assign 
            'UserConfig.cbs_topic_name' before 
            accessing it""",
    ):
        context.get_topic_name

def test_set_se_streaming_stats_kafka_custom_config_enable():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_se_streaming_stats_kafka_custom_config_enable(True)
    assert context.get_se_streaming_stats_kafka_custom_config_enable is True

def test_get_se_streaming_stats_kafka_custom_config_enable_default():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    assert context.get_se_streaming_stats_kafka_custom_config_enable is False


def test_get_se_streaming_stats_kafka_bootstrap_server():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_se_streaming_stats_kafka_bootstrap_server("test-kafka-server")
    assert context.get_se_streaming_stats_kafka_bootstrap_server == "test-kafka-server"


def test_get_se_streaming_stats_kafka_bootstrap_server_exception():
    context = SparkExpectationsContext(product_id="product1", spark=spark)

    with pytest.raises(
        SparkExpectationsMiscException,
        match="""The spark expectations context is not set completely, please assign 
            'se_streaming_stats_kafka_bootstrap_server' before 
            accessing it""",
    ):
        context.get_se_streaming_stats_kafka_bootstrap_server


def test_set_rules_exceeds_threshold():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_rules_exceeds_threshold(
        [
            {
                "rule_name": "rule_1",
                "action_if_failed": "ignore",
                "description": "description1",
                "rule_type": "row_dq",
                "error_drop_threshold": "10",
                "error_drop_percentage": "10.0",
            }
        ]
    )
    assert context.get_rules_exceeds_threshold == [
        {
            "rule_name": "rule_1",
            "action_if_failed": "ignore",
            "description": "description1",
            "rule_type": "row_dq",
            "error_drop_threshold": "10",
            "error_drop_percentage": "10.0",
        }
    ]


def test_get_rules_exceds_threshold():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context._rules_error_per = [
        {
            "rule_name": "rule_1",
            "action_if_failed": "ignore",
            "description": "description1",
            "rule_type": "row_dq",
            "error_drop_threshold": "10",
            "error_drop_percentage": "10.0",
        }
    ]

    assert context.get_rules_exceeds_threshold == [
        {
            "rule_name": "rule_1",
            "action_if_failed": "ignore",
            "description": "description1",
            "rule_type": "row_dq",
            "error_drop_threshold": "10",
            "error_drop_percentage": "10.0",
        }
    ]


def test_set_agg_dq_detailed_stats_status():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_agg_dq_detailed_stats_status(True)
    assert context.get_agg_dq_detailed_stats_status is True


def test_get_agg_dq_detailed_stats_status():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    assert context.get_agg_dq_detailed_stats_status is True


def test_set_query_dq_detailed_stats_status():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_query_dq_detailed_stats_status(True)
    assert context.get_query_dq_detailed_stats_status is True


def test_get_query_dq_detailed_stats_status():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    assert context.get_query_dq_detailed_stats_status is False


def test_set_target_agg_dq_detailed_stats():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    target_agg_dq_detailed_stats = [
        ("column1", "rule1"),
        ("column2", "rule2"),
        ("column3", "rule3"),
    ]
    context.set_target_agg_dq_detailed_stats(target_agg_dq_detailed_stats)
    assert context.get_target_agg_dq_detailed_stats == target_agg_dq_detailed_stats


def test_set_target_query_dq_detailed_stats():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    target_query_dq_detailed_stats = [
        ("column1", "rule1"),
        ("column2", "rule2"),
        ("column3", "rule3"),
    ]
    context.set_target_query_dq_detailed_stats(target_query_dq_detailed_stats)
    assert context.get_target_query_dq_detailed_stats == target_query_dq_detailed_stats


def test_set_source_agg_dq_detailed_stats():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    source_agg_dq_detailed_stats = [
        ("column1", "rule1"),
        ("column2", "rule2"),
        ("column3", "rule3"),
    ]
    context.set_source_agg_dq_detailed_stats(source_agg_dq_detailed_stats)
    assert context.get_source_agg_dq_detailed_stats == source_agg_dq_detailed_stats


def test_set_source_query_dq_detailed_stats():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    source_query_dq_detailed_stats = [
        ("column1", "rule1"),
        ("column2", "rule2"),
        ("column3", "rule3"),
    ]
    context.set_source_query_dq_detailed_stats(source_query_dq_detailed_stats)
    assert context.get_source_query_dq_detailed_stats == source_query_dq_detailed_stats


def test_set_detailed_stats_table_writer_config():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    config = {"key": "value"}
    context.set_detailed_stats_table_writer_config(config)
    assert context.get_detailed_stats_table_writer_config() == config


# Add more test cases as needed
def test_set_agg_dq_detailed_stats_status():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_agg_dq_detailed_stats_status(True)
    assert context.get_agg_dq_detailed_stats_status is True


def test_get_agg_dq_detailed_stats_status():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_agg_dq_detailed_stats_status(True)
    assert context.get_agg_dq_detailed_stats_status is True


def test_set_query_dq_detailed_stats_status():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_query_dq_detailed_stats_status(True)
    assert context.get_query_dq_detailed_stats_status is True


def test_get_query_dq_detailed_stats_status():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_query_dq_detailed_stats_status(True)
    assert context.get_query_dq_detailed_stats_status is True


def test_set_agg_dq_detailed_stats_status():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_agg_dq_detailed_stats_status(True)
    assert context.get_agg_dq_detailed_stats_status is True


def test_set_query_dq_detailed_stats_status():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_query_dq_detailed_stats_status(True)
    assert context.get_query_dq_detailed_stats_status is True


def test_set_dq_stats_table_name():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_dq_stats_table_name("dq_stats_table_name")
    assert context._dq_stats_table_name == "dq_stats_table_name"
    assert context.get_dq_stats_table_name == "dq_stats_table_name"


def test_set_dq_detailed_stats_table_name():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_dq_detailed_stats_table_name("dq_stats_table_name")

    context._enable_agg_dq_detailed_result = True
    context._enable_query_dq_detailed_result = True

    assert context.get_agg_dq_detailed_stats_status is True
    assert context.get_query_dq_detailed_stats_status is True

    assert context._dq_detailed_stats_table_name == "dq_stats_table_name"
    assert context.get_dq_detailed_stats_table_name == "dq_stats_table_name"


def test_set_detailed_stats_table_writer_config():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    config = {"setting1": "value1", "setting2": "value2"}
    context.set_detailed_stats_table_writer_config(config)
    assert context.get_detailed_stats_table_writer_config == config


def test_set_dq_expectations():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    dq_expectations = {"column1": "expectation1", "column2": "expectation2"}
    context.set_dq_expectations(dq_expectations)
    assert context.get_dq_expectations == dq_expectations


def test_set_rules_execution_settings_config():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    config = {"setting1": "value1", "setting2": "value2"}
    context.set_rules_execution_settings_config(config)
    assert context.get_rules_execution_settings_config == config


def test_get_row_dq_start_time():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context._row_dq_start_time = datetime.now()
    assert context.get_row_dq_start_time == context._row_dq_start_time


def test_get_row_dq_end_time():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context._row_dq_end_time = datetime.now()
    assert context.get_row_dq_end_time == context._row_dq_end_time


def test_get_row_dq_start_time_exception():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context._row_dq_start_time = None
    with pytest.raises(
        SparkExpectationsMiscException,
        match="""The spark expectations context is not set completely, 
            please assign '_row_dq_start_time'  before 
            accessing it""",
    ):
        context.get_row_dq_start_time


def test_get_row_dq_end_time_exception():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context._row_dq_end_time = None
    with pytest.raises(
        SparkExpectationsMiscException,
        match="""The spark expectations context is not set completely, please assign '_row_dq_end_time'  before 
            accessing it""",
    ):
        context.get_row_dq_end_time


def test_set_query_dq_detailed_stats_status():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_query_dq_detailed_stats_status(True)
    assert context._enable_query_dq_detailed_result == True
    assert context.get_query_dq_detailed_stats_status == True


def test_get_query_dq_output_custom_table_name():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_query_dq_detailed_stats_status(True)
    context.set_dq_detailed_stats_table_name("test_table")
    context.set_query_dq_output_custom_table_name("test_table")
    assert context.get_query_dq_output_custom_table_name == "test_table"


def test_get_query_dq_output_custom_table_name_exception():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_query_dq_detailed_stats_status(True)
    context.set_dq_detailed_stats_table_name(None)
    with pytest.raises(
        SparkExpectationsMiscException,
        match="""The spark expectations context is not set completely, please assign 
            '_dq_detailed_stats_table_name,query_dq_detailed_stats_status' before 
            accessing it""",
    ):
        context.get_query_dq_output_custom_table_name


def test_get_dq_detailed_stats_table_name():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_query_dq_detailed_stats_status(True)
    context.set_dq_detailed_stats_table_name("test_table")
    assert context.get_dq_detailed_stats_table_name == "test_table"


def test_get_dq_detailed_stats_table_name_exception():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_query_dq_detailed_stats_status(True)
    context.set_dq_detailed_stats_table_name(None)
    with pytest.raises(
        SparkExpectationsMiscException,
        match="""The spark expectations context is not set completely, please assign 
            '_dq_detailed_stats_table_name' before 
            accessing it""",
    ):
        context.get_dq_detailed_stats_table_name


def test_get_dq_expectations():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    dq_expectations = {"column1": "expectation1", "column2": "expectation2"}
    context.set_dq_expectations(dq_expectations)
    assert context.get_dq_expectations == dq_expectations


def test_get_dq_expectations_exception():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context._dq_expectations = None
    with pytest.raises(
        SparkExpectationsMiscException,
        match="""The spark expectations context is not set completely, please assign '_dq_expectations' before 
            accessing it""",
    ):
        context.get_dq_expectations


def test_set_querydq_secondary_queries():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    querydq_secondary_queries = {"query1": "query1", "query2": "query2"}
    context.set_querydq_secondary_queries(querydq_secondary_queries)
    assert context.get_querydq_secondary_queries == querydq_secondary_queries


def test_get_querydq_secondary_queries():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    querydq_secondary_queries = {"query1": "query1", "query2": "query2"}
    context.set_querydq_secondary_queries(querydq_secondary_queries)
    assert context.get_querydq_secondary_queries == querydq_secondary_queries


def test_get_source_query_dq_output():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    source_query_dq_output = [{"column1": "value1", "column2": "value2"}]
    context._source_query_dq_output = source_query_dq_output
    assert context.get_source_query_dq_output == source_query_dq_output


def test_set_target_query_dq_output():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    target_query_dq_output = [{"column1": "value1", "column2": "value2"}]
    context.set_target_query_dq_output(target_query_dq_output)
    assert context._target_query_dq_output == target_query_dq_output


def test_get_target_query_dq_output():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    target_query_dq_output = [{"column1": "value1", "column2": "value2"}]
    context._target_query_dq_output = target_query_dq_output
    assert context.get_target_query_dq_output == target_query_dq_output


def test_get_source_query_dq_output():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    source_query_dq_output = [{"column1": "value1", "column2": "value2"}]
    context._source_query_dq_output = source_query_dq_output
    assert context.get_source_query_dq_output == source_query_dq_output


def test_set_source_query_dq_output():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    source_query_dq_output = [{"column1": "value1", "column2": "value2"}]
    context.set_source_query_dq_output(source_query_dq_output)
    assert context._source_query_dq_output == source_query_dq_output


def test_set_enable_error_table():
    # default case is True for enabling error table
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    assert context.get_se_enable_error_table is True

    # testing for False do not write error records in error table
    context.set_se_enable_error_table(False)
    assert context.get_se_enable_error_table is False


def test_set_dq_rules_params():
    # default case is empty dictionary for dq rules params and testing negative scenario
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    assert context.get_dq_rules_params == {}

    # testing when passing parameterizied values to dq rules
    context._dq_rules_params = {"env": "local"}
    assert context.get_dq_rules_params == {"env": "local"}


def test_get_dq_expectations():
    context = SparkExpectationsContext(product_id="test_product", spark=spark)
    context.set_dq_expectations({"rule1": "expectation1", "rule2": "expectation2"})

    # Test when _dq_expectations is set
    assert context.get_dq_expectations == {
        "rule1": "expectation1",
        "rule2": "expectation2",
    }

    # Test when _dq_expectations is not set
    context._dq_expectations = None
    with pytest.raises(SparkExpectationsMiscException):
        context.get_dq_expectations()


def test_set_dq_expectations():
    context = SparkExpectationsContext(product_id="test_product", spark=spark)
    dq_expectations = {
        "rule1": "expectation1",
        "rule2": "expectation2",
        "rule3": "expectation3",
    }
    context.set_dq_expectations(dq_expectations)
    assert context._dq_expectations == dq_expectations


def test_set_job_metadata():
    context = SparkExpectationsContext(product_id="test_product", spark=spark)
    context.set_job_metadata("{'job_name': 'test_job_metadata'}")
    assert context._job_metadata == "{'job_name': 'test_job_metadata'}"


def test_get_job_metadata():
    context = SparkExpectationsContext(product_id="test_product", spark=spark)
    context._job_metadata = "{'job_name': 'test_job_metadata'}"
    assert context.get_job_metadata == "{'job_name': 'test_job_metadata'}"

    # testing for None condition
    context._job_metadata = None
    assert context.get_job_metadata is None


def test_set_enable_obs_dq_report_result():
    context = SparkExpectationsContext(product_id="test_product", spark=spark)
    context.set_enable_obs_dq_report_result(True)
    assert context.get_enable_obs_dq_report_result is True


def test_get_enable_obs_dq_report_result():
    context = SparkExpectationsContext(product_id="test_product", spark=spark)
    context.set_enable_obs_dq_report_result(True)
    assert context.get_enable_obs_dq_report_result is True

    # testing for None condition
    context._enable_obs_dq_report_result = None
    assert context.get_enable_obs_dq_report_result is None


def test_set_se_dq_obs_alert_flag():
    context = SparkExpectationsContext(product_id="test_product", spark=spark)
    context.set_se_dq_obs_alert_flag(True)
    assert context.get_se_dq_obs_alert_flag is True


def test_get_se_dq_obs_alert_flag():
    context = SparkExpectationsContext(product_id="test_product", spark=spark)
    context.set_se_dq_obs_alert_flag(True)
    assert context.get_se_dq_obs_alert_flag is True

    # testing for None condition
    context._se_dq_obs_alert_flag = None
    assert context.get_se_dq_obs_alert_flag is None


def test_set_detailed_default_template():
    context = SparkExpectationsContext(product_id="test_product", spark=spark)
    context.set_detailed_default_template("test_template")
    assert context.get_detailed_default_template == "test_template"


def test_get_detailed_default_template():
    context = SparkExpectationsContext(product_id="test_product", spark=spark)
    context.set_detailed_default_template("test_template")
    assert context.get_detailed_default_template == "test_template"

    # testing for None condition
    context._detailed_default_template = None
    assert context.get_detailed_default_template is None


def test_set_basic_default_template():
    context = SparkExpectationsContext(product_id="test_product", spark=spark)
    context.set_basic_default_template("test_template")
    assert context.get_basic_default_template == "test_template"


def test_get_basic_default_template():
    context = SparkExpectationsContext(product_id="test_product", spark=spark)
    context.set_basic_default_template("test_template")
    assert context.get_basic_default_template == "test_template"

    # testing for None condition
    context._basic_default_template = None
    assert context.get_basic_default_template is None


def test_set_custom_default_template():
    context = SparkExpectationsContext(product_id="test_product", spark=spark)
    context.set_custom_default_template("test_template")
    assert context.get_custom_default_template == "test_template"


def test_get_custom_default_template():
    context = SparkExpectationsContext(product_id="test_product", spark=spark)
    context.set_custom_default_template("test_template")
    assert context.get_custom_default_template == "test_template"

    # testing for None condition
    context._custom_default_template = None
    assert context.get_custom_default_template is None


def test_set_stats_detailed_dataframe():
    context = SparkExpectationsContext(product_id="test_product", spark=spark)
    data = [("test_product_id", "test_schema.table1", 5)]
    columns = ["product_id", "table_name", "error_count"]
    df = spark.createDataFrame(data, columns)
    context.set_stats_detailed_dataframe(df)


def test_get_stats_detailed_dataframe():
    context = SparkExpectationsContext(product_id="test_product", spark=spark)
    data = [("test_product_id", "test_schema.table1", 5)]
    columns = ["product_id", "table_name", "error_count"]
    df = spark.createDataFrame(data, columns)
    context._stats_detailed_dataframe = df

    if context.get_stats_detailed_dataframe is None:
        assert context.get_stats_detailed_dataframe is None
    else:
        assert context.get_stats_detailed_dataframe.collect() == df.collect()


def test_set_custom_detailed_dataframe():
    context = SparkExpectationsContext(product_id="test_product", spark=spark)
    data = [("test_product_id", "test_schema.table1", 5)]
    columns = ["product_id", "table_name", "error_count"]
    df = spark.createDataFrame(data, columns)
    context.set_custom_detailed_dataframe(df)


def test_get_custom_detailed_dataframe():
    context = SparkExpectationsContext(product_id="test_product", spark=spark)
    data = [("test_product_id", "test_schema.table1", 5)]
    columns = ["product_id", "table_name", "error_count"]
    df = spark.createDataFrame(data, columns)
    context._custom_detailed_dataframe = df

    if context.get_custom_detailed_dataframe is None:
        assert context.get_custom_detailed_dataframe is None
    else:
        assert context.get_custom_detailed_dataframe.collect() == df.collect()


def test_set_report_table_name():
    context = SparkExpectationsContext(product_id="test_product", spark=spark)
    context.set_report_table_name("test_table")
    assert context.get_report_table_name == "test_table"


def test_get_report_table_name():
    context = SparkExpectationsContext(product_id="test_product", spark=spark)
    context.set_report_table_name("test_table")
    assert context.get_report_table_name == "test_table"

    # testing for None condition
    context._report_table_name = None
    assert context.get_report_table_name is None


def test_set_dq_obs_rpt_gen_status_flag():
    context = SparkExpectationsContext(product_id="test_product", spark=spark)
    context.set_dq_obs_rpt_gen_status_flag(True)
    assert context.get_dq_obs_rpt_gen_status_flag is True


def test_get_dq_obs_rpt_gen_status_flag():
    context = SparkExpectationsContext(product_id="test_product", spark=spark)
    context.set_dq_obs_rpt_gen_status_flag(True)
    assert context.get_dq_obs_rpt_gen_status_flag is True

    # testing for None condition
    context._dq_obs_rpt_gen_status_flag = None
    assert context.get_dq_obs_rpt_gen_status_flag is None


def test_set_df_dq_obs_report_dataframe():
    #
    context = SparkExpectationsContext(product_id="test_product", spark=spark)
    data = [("test_product_id", "test_schema.table1", 5)]
    columns = ["product_id", "table_name", "error_count"]
    df = spark.createDataFrame(data, columns)
    context.set_df_dq_obs_report_dataframe(df)


def test_get_df_dq_obs_report_dataframe():
    context = SparkExpectationsContext(product_id="test_product", spark=spark)
    data = [("test_product_id", "test_schema.table1", 5)]
    columns = ["product_id", "table_name", "error_count"]
    df = spark.createDataFrame(data, columns)
    context._df_dq_obs_report_dataframe = df

    if context.get_df_dq_obs_report_dataframe is None:
        assert context.get_df_dq_obs_report_dataframe is None
    else:
        assert context.get_df_dq_obs_report_dataframe.collect() == df.collect()


# [Unit Tests for writer type and config methods]


def test_set_target_and_error_table_writer_type():
    """Test setting target and error table writer type"""
    context = SparkExpectationsContext(
        product_id="product1",
        spark=spark,
    )
    context.set_target_and_error_table_writer_type("streaming")
    assert context.get_target_and_error_table_writer_type == "streaming"


def test_get_target_and_error_table_writer_type_default():
    """Test getting default target and error table writer type"""
    context = SparkExpectationsContext(
        product_id="product1",
        spark=spark,
    )
    # Default should be "batch"
    assert context.get_target_and_error_table_writer_type == "batch"


def test_set_stats_table_writer_type():
    """Test setting stats table writer type"""
    context = SparkExpectationsContext(
        product_id="product1",
        spark=spark,
    )
    context.set_stats_table_writer_type("streaming")
    assert context.get_stats_table_writer_type == "streaming"


def test_get_stats_table_writer_type_default():
    """Test getting default stats table writer type"""
    context = SparkExpectationsContext(
        product_id="product1",
        spark=spark,
    )
    # Default should be "batch"
    assert context.get_stats_table_writer_type == "batch"


def test_set_target_and_error_table_writer_config():
    """Test setting target and error table writer config"""
    context = SparkExpectationsContext(
        product_id="product1",
        spark=spark,
    )
    config = {
        "mode": "append",
        "format": "delta",
        "partitionBy": ["date"],
        "options": {"checkpointLocation": "/path/to/checkpoint"},
    }
    context.set_target_and_error_table_writer_config(config)
    assert context.get_target_and_error_table_writer_config == config


def test_get_target_and_error_table_writer_config_default():
    """Test getting default target and error table writer config"""
    context = SparkExpectationsContext(
        product_id="product1",
        spark=spark,
    )
    # Default should be empty dict
    assert context.get_target_and_error_table_writer_config == {}


def test_set_stats_table_writer_config():
    """Test setting stats table writer config"""
    context = SparkExpectationsContext(
        product_id="product1",
        spark=spark,
    )
    config = {
        "outputMode": "append",
        "format": "delta",
        "queryName": "test_query",
        "trigger": {"processingTime": "10 seconds"},
        "partitionBy": ["date"],
        "options": {"checkpointLocation": "/path/to/checkpoint"},
    }
    context.set_stats_table_writer_config(config)
    assert context.get_stats_table_writer_config == config


def test_get_stats_table_writer_config_default():
    """Test getting default stats table writer config"""
    context = SparkExpectationsContext(
        product_id="product1",
        spark=spark,
    )
    # Default should be empty dict
    assert context.get_stats_table_writer_config == {}


def test_writer_type_batch_to_streaming_transition():
    """Test transitioning writer type from batch to streaming"""
    context = SparkExpectationsContext(
        product_id="product1",
        spark=spark,
    )
    # Default is batch
    assert context.get_target_and_error_table_writer_type == "batch"
    assert context.get_stats_table_writer_type == "batch"
    
    # Change to streaming
    context.set_target_and_error_table_writer_type("streaming")
    context.set_stats_table_writer_type("streaming")
    
    assert context.get_target_and_error_table_writer_type == "streaming"
    assert context.get_stats_table_writer_type == "streaming"


def test_writer_config_update():
    """Test updating writer config"""
    context = SparkExpectationsContext(
        product_id="product1",
        spark=spark,
    )
    
    # Set initial config
    initial_config = {
        "mode": "append",
        "format": "delta",
    }
    context.set_target_and_error_table_writer_config(initial_config)
    assert context.get_target_and_error_table_writer_config == initial_config
    
    # Update config
    updated_config = {
        "mode": "overwrite",
        "format": "parquet",
        "partitionBy": ["date", "region"],
    }
    context.set_target_and_error_table_writer_config(updated_config)
    assert context.get_target_and_error_table_writer_config == updated_config


def test_streaming_writer_config():
    """Test setting streaming writer config with all options"""
    context = SparkExpectationsContext(
        product_id="product1",
        spark=spark,
    )
    
    streaming_config = {
        "outputMode": "append",
        "format": "delta",
        "queryName": "my_streaming_query",
        "trigger": {"processingTime": "10 seconds"},
        "partitionBy": ["date", "hour"],
        "options": {
            "checkpointLocation": "/path/to/checkpoint",
            "maxFilesPerTrigger": "100",
        },
    }
    
    context.set_stats_table_writer_config(streaming_config)
    retrieved_config = context.get_stats_table_writer_config
    
    assert retrieved_config["outputMode"] == "append"
    assert retrieved_config["format"] == "delta"
    assert retrieved_config["queryName"] == "my_streaming_query"
    assert retrieved_config["trigger"]["processingTime"] == "10 seconds"
    assert retrieved_config["partitionBy"] == ["date", "hour"]
    assert retrieved_config["options"]["checkpointLocation"] == "/path/to/checkpoint"
    assert retrieved_config["options"]["maxFilesPerTrigger"] == "100"
def test_set_min_priority_email():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_min_priority_email("high")
    assert context._min_priority_email == "high"
    assert context.get_min_priority_email == "high"


def test_get_min_priority_email():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_min_priority_email("low")
    assert context.get_min_priority_email == "low"  # default value


def test_set_min_priority_pagerduty():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_min_priority_pagerduty("medium")
    assert context._min_priority_pagerduty == "medium"
    assert context.get_min_priority_pagerduty == "medium"


def test_get_min_priority_pagerduty():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_min_priority_pagerduty("low")
    assert context.get_min_priority_pagerduty == "low"  # default value


def test_set_min_priority_slack():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_min_priority_slack("high")
    assert context._min_priority_slack == "high"
    assert context.get_min_priority_slack == "high"


def test_get_min_priority_slack():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_min_priority_slack("low")
    assert context.get_min_priority_slack == "low"  # default value


def test_set_min_priority_teams():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_min_priority_teams("medium")
    assert context._min_priority_teams == "medium"
    assert context.get_min_priority_teams == "medium"


def test_get_min_priority_teams():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_min_priority_teams("low")
    assert context.get_min_priority_teams == "low"  # default value


def test_set_min_priority_zoom():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_min_priority_zoom("high")
    assert context._min_priority_zoom == "high"
    assert context.get_min_priority_zoom == "high"


def test_get_min_priority_zoom():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_min_priority_zoom("low")
    assert context.get_min_priority_zoom == "low"  # default value


def test_set_kafka_write_status():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_kafka_write_status("Success")
    assert context._kafka_write_status == "Success"
    assert context.get_kafka_write_status == "Success"


def test_get_kafka_write_status():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    context.set_kafka_write_status("Failed")
    assert context.get_kafka_write_status == "Failed"


def test_get_kafka_write_status_default():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    # Test default value when attribute not set
    assert context.get_kafka_write_status == "Disabled"


def test_set_kafka_write_error_message():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    error_msg = "Connection refused: kafka broker not available"
    context.set_kafka_write_error_message(error_msg)
    assert context._kafka_write_error_message == error_msg
    assert context.get_kafka_write_error_message == error_msg


def test_get_kafka_write_error_message():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    error_msg = "Kafka write timeout"
    context.set_kafka_write_error_message(error_msg)
    assert context.get_kafka_write_error_message == error_msg


def test_get_kafka_write_error_message_default():
    context = SparkExpectationsContext(product_id="product1", spark=spark)
    # Test default value when attribute not set
    assert context.get_kafka_write_error_message == ""



# c = get_config()

# # Set password (replace 'your_password' with your actual password)
# from notebook.auth import passwd
# c.ServerApp.password = passwd("admin")

# # Disable token authentication
# c.ServerApp.token = ''

# Configuration file for lab.

c = get_config()  # noqa

#------------------------------------------------------------------------------
# Application(SingletonConfigurable) configuration
#------------------------------------------------------------------------------

# Set default theme
# c.LabApp.theme = "JupyterLab Dark" 

## This is an application.

## The date format used by logging formatters for %(asctime)s
#  Default: '%Y-%m-%d %H:%M:%S'
# c.Application.log_datefmt = '%Y-%m-%d %H:%M:%S'

## The Logging format template
#  Default: '[%(name)s]%(highlevel)s %(message)s'
# c.Application.log_format = '[%(name)s]%(highlevel)s %(message)s'

## Set the log level by value or name.
#  Choices: any of [0, 10, 20, 30, 40, 50, 'DEBUG', 'INFO', 'WARN', 'ERROR', 'CRITICAL']
#  Default: 30
# c.Application.log_level = 30

## Configure additional log handlers.
#  
#  The default stderr logs handler is configured by the log_level, log_datefmt
#  and log_format settings.
#  
#  This configuration can be used to configure additional handlers (e.g. to
#  output the log to a file) or for finer control over the default handlers.
#  
#  If provided this should be a logging configuration dictionary, for more
#  information see:
#  https://docs.python.org/3/library/logging.config.html#logging-config-
#  dictschema
#  
#  This dictionary is merged with the base logging configuration which defines
#  the following:
#  
#  * A logging formatter intended for interactive use called
#    ``console``.
#  * A logging handler that writes to stderr called
#    ``console`` which uses the formatter ``console``.
#  * A logger with the name of this application set to ``DEBUG``
#    level.
#  
#  This example adds a new handler that writes to a file:
#  
#  .. code-block:: python
#  
#     c.Application.logging_config = {
#         "handlers": {
#             "file": {
#                 "class": "logging.FileHandler",
#                 "level": "DEBUG",
#                 "filename": "<path/to/file>",
#             }
#         },
#         "loggers": {
#             "<application-name>": {
#                 "level": "DEBUG",
#                 # NOTE: if you don't list the default "console"
#                 # handler here then it will be disabled
#                 "handlers": ["console", "file"],
#             },
#         },
#     }
#  Default: {}
# c.Application.logging_config = {}

## Instead of starting the Application, dump configuration to stdout
#  Default: False
# c.Application.show_config = False

## Instead of starting the Application, dump configuration to stdout (as JSON)
#  Default: False
# c.Application.show_config_json = False

#------------------------------------------------------------------------------
# JupyterApp(Application) configuration
#------------------------------------------------------------------------------
## Base class for Jupyter applications

## Answer yes to any prompts.
#  Default: False
# c.JupyterApp.answer_yes = False

## Full path of a config file.
#  Default: ''
# c.JupyterApp.config_file = ''

## Specify a config file to load.
#  Default: ''
# c.JupyterApp.config_file_name = ''

## Generate default config file.
#  Default: False
# c.JupyterApp.generate_config = False

## The date format used by logging formatters for %(asctime)s
#  See also: Application.log_datefmt
# c.JupyterApp.log_datefmt = '%Y-%m-%d %H:%M:%S'

## The Logging format template
#  See also: Application.log_format
# c.JupyterApp.log_format = '[%(name)s]%(highlevel)s %(message)s'

## Set the log level by value or name.
#  See also: Application.log_level
# c.JupyterApp.log_level = 30

## 
#  See also: Application.logging_config
# c.JupyterApp.logging_config = {}

## Instead of starting the Application, dump configuration to stdout
#  See also: Application.show_config
# c.JupyterApp.show_config = False

## Instead of starting the Application, dump configuration to stdout (as JSON)
#  See also: Application.show_config_json
# c.JupyterApp.show_config_json = False

#------------------------------------------------------------------------------
# ExtensionApp(JupyterApp) configuration
#------------------------------------------------------------------------------
## Base class for configurable Jupyter Server Extension Applications.
#  
#  ExtensionApp subclasses can be initialized two ways:
#  
#  - Extension is listed as a jpserver_extension, and ServerApp calls
#    its load_jupyter_server_extension classmethod. This is the
#    classic way of loading a server extension.
#  
#  - Extension is launched directly by calling its `launch_instance`
#    class method. This method can be set as a entry_point in
#    the extensions setup.py.

## Answer yes to any prompts.
#  See also: JupyterApp.answer_yes
# c.ExtensionApp.answer_yes = False

## Full path of a config file.
#  See also: JupyterApp.config_file
# c.ExtensionApp.config_file = ''

## Specify a config file to load.
#  See also: JupyterApp.config_file_name
# c.ExtensionApp.config_file_name = ''

#  Default: ''
# c.ExtensionApp.default_url = ''

## Generate default config file.
#  See also: JupyterApp.generate_config
# c.ExtensionApp.generate_config = False

## Handlers appended to the server.
#  Default: []
# c.ExtensionApp.handlers = []

## The date format used by logging formatters for %(asctime)s
#  See also: Application.log_datefmt
# c.ExtensionApp.log_datefmt = '%Y-%m-%d %H:%M:%S'

## The Logging format template
#  See also: Application.log_format
# c.ExtensionApp.log_format = '[%(name)s]%(highlevel)s %(message)s'

## Set the log level by value or name.
#  See also: Application.log_level
# c.ExtensionApp.log_level = 30

## 
#  See also: Application.logging_config
# c.ExtensionApp.logging_config = {}

## Whether to open in a browser after starting.
#          The specific browser used is platform dependent and
#          determined by the python standard library `webbrowser`
#          module, unless it is overridden using the --browser
#          (ServerApp.browser) configuration option.
#  Default: False
# c.ExtensionApp.open_browser = False

## Settings that will passed to the server.
#  Default: {}
# c.ExtensionApp.settings = {}

## Instead of starting the Application, dump configuration to stdout
#  See also: Application.show_config
# c.ExtensionApp.show_config = False

## Instead of starting the Application, dump configuration to stdout (as JSON)
#  See also: Application.show_config_json
# c.ExtensionApp.show_config_json = False

## paths to search for serving static files.
#  
#          This allows adding javascript/css to be available from the notebook server machine,
#          or overriding individual files in the IPython
#  Default: []
# c.ExtensionApp.static_paths = []

## Url where the static assets for the extension are served.
#  Default: ''
# c.ExtensionApp.static_url_prefix = ''

## Paths to search for serving jinja templates.
#  
#          Can be used to override templates from notebook.templates.
#  Default: []
# c.ExtensionApp.template_paths = []

#------------------------------------------------------------------------------
# LabServerApp(ExtensionApp) configuration
#------------------------------------------------------------------------------
## A Lab Server Application that runs out-of-the-box

## "A list of comma-separated URIs to get the allowed extensions list
#  
#  .. versionchanged:: 2.0.0
#      `LabServerApp.whitetlist_uris` renamed to `allowed_extensions_uris`
#  Default: ''
# c.LabServerApp.allowed_extensions_uris = ''

## Answer yes to any prompts.
#  See also: JupyterApp.answer_yes
# c.LabServerApp.answer_yes = False

## The application settings directory.
#  Default: ''
# c.LabServerApp.app_settings_dir = ''

## The url path for the application.
#  Default: '/lab'
# c.LabServerApp.app_url = '/lab'

## Deprecated, use `LabServerApp.blocked_extensions_uris`
#  Default: ''
# c.LabServerApp.blacklist_uris = ''

## A list of comma-separated URIs to get the blocked extensions list
#  
#  .. versionchanged:: 2.0.0
#      `LabServerApp.blacklist_uris` renamed to `blocked_extensions_uris`
#  Default: ''
# c.LabServerApp.blocked_extensions_uris = ''

## Whether to cache files on the server. This should be `True` except in dev
#  mode.
#  Default: True
# c.LabServerApp.cache_files = True

## Full path of a config file.
#  See also: JupyterApp.config_file
# c.LabServerApp.config_file = ''

## Specify a config file to load.
#  See also: JupyterApp.config_file_name
# c.LabServerApp.config_file_name = ''

## Whether getting a relative (False) or absolute (True) path when copying a
#  path.
#  Default: False
# c.LabServerApp.copy_absolute_path = False

## Extra paths to look for federated JupyterLab extensions
#  Default: []
# c.LabServerApp.extra_labextensions_path = []

## Generate default config file.
#  See also: JupyterApp.generate_config
# c.LabServerApp.generate_config = False

## Handlers appended to the server.
#  See also: ExtensionApp.handlers
# c.LabServerApp.handlers = []

## Options to pass to the jinja2 environment for this
#  Default: {}
# c.LabServerApp.jinja2_options = {}

## The standard paths to look in for federated JupyterLab extensions
#  Default: []
# c.LabServerApp.labextensions_path = []

## The url for federated JupyterLab extensions
#  Default: ''
# c.LabServerApp.labextensions_url = ''

## The interval delay in seconds to refresh the lists
#  Default: 3600
# c.LabServerApp.listings_refresh_seconds = 3600

## The optional kwargs to use for the listings HTTP requests             as
#  described on https://2.python-requests.org/en/v2.7.0/api/#requests.request
#  Default: {}
# c.LabServerApp.listings_request_options = {}

## The listings url.
#  Default: ''
# c.LabServerApp.listings_url = ''

## The date format used by logging formatters for %(asctime)s
#  See also: Application.log_datefmt
# c.LabServerApp.log_datefmt = '%Y-%m-%d %H:%M:%S'

## The Logging format template
#  See also: Application.log_format
# c.LabServerApp.log_format = '[%(name)s]%(highlevel)s %(message)s'

## Set the log level by value or name.
#  See also: Application.log_level
# c.LabServerApp.log_level = 30

## 
#  See also: Application.logging_config
# c.LabServerApp.logging_config = {}

## Whether a notebook should start a kernel automatically.
#  Default: True
# c.LabServerApp.notebook_starts_kernel = True

## Whether to open in a browser after starting.
#  See also: ExtensionApp.open_browser
# c.LabServerApp.open_browser = False

## The optional location of the settings schemas directory. If given, a handler
#  will be added for settings.
#  Default: ''
# c.LabServerApp.schemas_dir = ''

## Settings that will passed to the server.
#  See also: ExtensionApp.settings
# c.LabServerApp.settings = {}

## The url path of the settings handler.
#  Default: ''
# c.LabServerApp.settings_url = ''

## Instead of starting the Application, dump configuration to stdout
#  See also: Application.show_config
# c.LabServerApp.show_config = False

## Instead of starting the Application, dump configuration to stdout (as JSON)
#  See also: Application.show_config_json
# c.LabServerApp.show_config_json = False

## The optional location of local static files. If given, a static file handler
#  will be added.
#  Default: ''
# c.LabServerApp.static_dir = ''

## paths to search for serving static files.
#  See also: ExtensionApp.static_paths
# c.LabServerApp.static_paths = []

## Url where the static assets for the extension are served.
#  See also: ExtensionApp.static_url_prefix
# c.LabServerApp.static_url_prefix = ''

## Paths to search for serving jinja templates.
#  See also: ExtensionApp.template_paths
# c.LabServerApp.template_paths = []

## The application templates directory.
#  Default: ''
# c.LabServerApp.templates_dir = ''

## The optional location of the themes directory. If given, a handler will be
#  added for themes.
#  Default: ''
# c.LabServerApp.themes_dir = ''

## The theme url.
#  Default: ''
# c.LabServerApp.themes_url = ''

## The url path of the translations handler.
#  Default: ''
# c.LabServerApp.translations_api_url = ''

## The url path of the tree handler.
#  Default: ''
# c.LabServerApp.tree_url = ''

## The optional location of the user settings directory.
#  Default: ''
# c.LabServerApp.user_settings_dir = ''

## Deprecated, use `LabServerApp.allowed_extensions_uris`
#  Default: ''
# c.LabServerApp.whitelist_uris = ''

## The url path of the workspaces API.
#  Default: ''
# c.LabServerApp.workspaces_api_url = ''

## The optional location of the saved workspaces directory. If given, a handler
#  will be added for workspaces.
#  Default: ''
# c.LabServerApp.workspaces_dir = ''

#------------------------------------------------------------------------------
# LabApp(LabServerApp) configuration
#------------------------------------------------------------------------------
## 
#  See also: LabServerApp.allowed_extensions_uris
# c.LabApp.allowed_extensions_uris = ''

## Answer yes to any prompts.
#  See also: JupyterApp.answer_yes
# c.LabApp.answer_yes = False

## The app directory to launch JupyterLab from.
#  Default: None
# c.LabApp.app_dir = None

## The application settings directory.
#  Default: ''
# c.LabApp.app_settings_dir = ''

## The url path for the application.
#  Default: '/lab'
# c.LabApp.app_url = '/lab'

## Deprecated, use `LabServerApp.blocked_extensions_uris`
#  See also: LabServerApp.blacklist_uris
# c.LabApp.blacklist_uris = ''

## 
#  See also: LabServerApp.blocked_extensions_uris
# c.LabApp.blocked_extensions_uris = ''

## Whether to cache files on the server. This should be `True` except in dev
#  mode.
#  Default: True
# c.LabApp.cache_files = True

## A callable class that receives the current version at instantiation and
#  calling it must return asynchronously a string indicating which version is
#  available and how to install or None if no update is available. The string
#  supports Markdown format.
#  Default: 'jupyterlab.handlers.announcements.CheckForUpdate'
# c.LabApp.check_for_updates_class = 'jupyterlab.handlers.announcements.CheckForUpdate'

## To enable real-time collaboration, you must install the extension `jupyter_collaboration`.
#          You can install it using pip for example:
#  
#              python -m pip install jupyter_collaboration
#  
#          This flag is now deprecated and will be removed in JupyterLab v5.
#  Default: False
# c.LabApp.collaborative = False

## Full path of a config file.
#  See also: JupyterApp.config_file
# c.LabApp.config_file = ''

## Specify a config file to load.
#  See also: JupyterApp.config_file_name
# c.LabApp.config_file_name = ''

## Whether getting a relative (False) or absolute (True) path when copying a
#  path.
#  Default: False
# c.LabApp.copy_absolute_path = False

## Whether to start the app in core mode. In this mode, JupyterLab
#          will run using the JavaScript assets that are within the installed
#          JupyterLab Python package. In core mode, third party extensions are disabled.
#          The `--dev-mode` flag is an alias to this to be used when the Python package
#          itself is installed in development mode (`pip install -e .`).
#  Default: False
# c.LabApp.core_mode = False

## Whether custom CSS is loaded on the page.
#      Defaults to False.
#  Default: False
# c.LabApp.custom_css = False

## The default URL to redirect to from `/`
#  Default: '/lab'
# c.LabApp.default_url = '/lab'

## Whether to start the app in dev mode. Uses the unpublished local
#          JavaScript packages in the `dev_mode` folder.  In this case JupyterLab will
#          show a red stripe at the top of the page.  It can only be used if JupyterLab
#          is installed as `pip install -e .`.
#  Default: False
# c.LabApp.dev_mode = False

## Whether to expose the global app instance to browser via window.jupyterapp
#  Default: False
# c.LabApp.expose_app_in_browser = False

## The extension manager factory to use. The default options are:
#          "readonly" for a manager without installation capability or "pypi" for
#          a manager using PyPi.org and pip to install extensions.
#  Default: 'pypi'
# c.LabApp.extension_manager = 'pypi'

## Whether to load prebuilt extensions in dev mode. This may be
#          useful to run and test prebuilt extensions in development installs of
#          JupyterLab. APIs in a JupyterLab development install may be
#          incompatible with published packages, so prebuilt extensions compiled
#          against published packages may not work correctly.
#  Default: False
# c.LabApp.extensions_in_dev_mode = False

## Extra paths to look for federated JupyterLab extensions
#  Default: []
# c.LabApp.extra_labextensions_path = []

## Generate default config file.
#  See also: JupyterApp.generate_config
# c.LabApp.generate_config = False

## Handlers appended to the server.
#  See also: ExtensionApp.handlers
# c.LabApp.handlers = []

## Options to pass to the jinja2 environment for this
#  Default: {}
# c.LabApp.jinja2_options = {}

## The standard paths to look in for federated JupyterLab extensions
#  Default: []
# c.LabApp.labextensions_path = []

## The url for federated JupyterLab extensions
#  Default: ''
# c.LabApp.labextensions_url = ''

## The interval delay in seconds to refresh the lists
#  See also: LabServerApp.listings_refresh_seconds
# c.LabApp.listings_refresh_seconds = 3600

## The optional kwargs to use for the listings HTTP requests             as
#  described on https://2.python-requests.org/en/v2.7.0/api/#requests.request
#  See also: LabServerApp.listings_request_options
# c.LabApp.listings_request_options = {}

## The listings url.
#  Default: ''
# c.LabApp.listings_url = ''

## Whether all plugins are locked (cannot be enabled/disabled from the UI)
#  Default: False
# c.LabApp.lock_all_plugins = False

## The date format used by logging formatters for %(asctime)s
#  See also: Application.log_datefmt
# c.LabApp.log_datefmt = '%Y-%m-%d %H:%M:%S'

## The Logging format template
#  See also: Application.log_format
# c.LabApp.log_format = '[%(name)s]%(highlevel)s %(message)s'

## Set the log level by value or name.
#  See also: Application.log_level
# c.LabApp.log_level = 30

## 
#  See also: Application.logging_config
# c.LabApp.logging_config = {}

## URL that serves news Atom feed; by default the JupyterLab organization
#  announcements will be fetched. Set to None to turn off fetching announcements.
#  Default: 'https://jupyterlab.github.io/assets/feed.xml'
# c.LabApp.news_url = 'https://jupyterlab.github.io/assets/feed.xml'

## Whether a notebook should start a kernel automatically.
#  Default: True
# c.LabApp.notebook_starts_kernel = True

## Whether to open in a browser after starting.
#  See also: ExtensionApp.open_browser
# c.LabApp.open_browser = False

## The override url for static lab assets, typically a CDN.
#  Default: ''
# c.LabApp.override_static_url = ''

## The override url for static lab theme assets, typically a CDN.
#  Default: ''
# c.LabApp.override_theme_url = ''

## The optional location of the settings schemas directory. If given, a handler
#  will be added for settings.
#  Default: ''
# c.LabApp.schemas_dir = ''

## Settings that will passed to the server.
#  See also: ExtensionApp.settings
# c.LabApp.settings = {}

## The url path of the settings handler.
#  Default: ''
# c.LabApp.settings_url = ''

## Instead of starting the Application, dump configuration to stdout
#  See also: Application.show_config
# c.LabApp.show_config = False

## Instead of starting the Application, dump configuration to stdout (as JSON)
#  See also: Application.show_config_json
# c.LabApp.show_config_json = False

## Whether to skip the initial install and JS build of the app in dev mode
#  Default: False
# c.LabApp.skip_dev_build = False

## Splice source packages into app directory.
#  Default: False
# c.LabApp.splice_source = False

## The optional location of local static files. If given, a static file handler
#  will be added.
#  Default: ''
# c.LabApp.static_dir = ''

## paths to search for serving static files.
#  See also: ExtensionApp.static_paths
# c.LabApp.static_paths = []

## Url where the static assets for the extension are served.
#  See also: ExtensionApp.static_url_prefix
# c.LabApp.static_url_prefix = ''

## Paths to search for serving jinja templates.
#  See also: ExtensionApp.template_paths
# c.LabApp.template_paths = []

## The application templates directory.
#  Default: ''
# c.LabApp.templates_dir = ''

## The optional location of the themes directory. If given, a handler will be
#  added for themes.
#  Default: ''
# c.LabApp.themes_dir = ''

## The theme url.
#  Default: ''
# c.LabApp.themes_url = ''

## The url path of the translations handler.
#  Default: ''
# c.LabApp.translations_api_url = ''

## The url path of the tree handler.
#  Default: ''
# c.LabApp.tree_url = ''

## The directory for user settings.
#  Default: '/Users/SMirko/.jupyter/lab/user-settings'
# c.LabApp.user_settings_dir = '/Users/SMirko/.jupyter/lab/user-settings'

## Whether to serve the app in watch mode
#  Default: False
# c.LabApp.watch = False

## Deprecated, use `LabServerApp.allowed_extensions_uris`
#  See also: LabServerApp.whitelist_uris
# c.LabApp.whitelist_uris = ''

## The url path of the workspaces API.
#  Default: ''
# c.LabApp.workspaces_api_url = ''

## The directory for workspaces
#  Default: '/Users/SMirko/.jupyter/lab/workspaces'
# c.LabApp.workspaces_dir = '/Users/SMirko/.jupyter/lab/workspaces'

#------------------------------------------------------------------------------
# ServerApp(JupyterApp) configuration
#------------------------------------------------------------------------------
## The Jupyter Server application class.

## Set the Access-Control-Allow-Credentials: true header
#  Default: False
# c.ServerApp.allow_credentials = False

## Whether or not to allow external kernels, whose connection files are placed in
#  external_connection_dir.
#  Default: False
# c.ServerApp.allow_external_kernels = False

## Set the Access-Control-Allow-Origin header
#  
#          Use '*' to allow any origin to access your server.
#  
#          Takes precedence over allow_origin_pat.
#  Default: ''
c.ServerApp.allow_origin = '*'

## Use a regular expression for the Access-Control-Allow-Origin header
#  
#          Requests from an origin matching the expression will get replies with:
#  
#              Access-Control-Allow-Origin: origin
#  
#          where `origin` is the origin of the request.
#  
#          Ignored if allow_origin is set.
#  Default: ''
# c.ServerApp.allow_origin_pat = ''

## DEPRECATED in 2.0. Use PasswordIdentityProvider.allow_password_change
#  Default: True
# c.ServerApp.allow_password_change = True

## Allow requests where the Host header doesn't point to a local server
#  
#         By default, requests get a 403 forbidden response if the 'Host' header
#         shows that the browser thinks it's on a non-local domain.
#         Setting this option to True disables this check.
#  
#         This protects against 'DNS rebinding' attacks, where a remote web server
#         serves you a page and then changes its DNS to send later requests to a
#         local IP, bypassing same-origin checks.
#  
#         Local IP addresses (such as 127.0.0.1 and ::1) are allowed as local,
#         along with hostnames configured in local_hostnames.
#  Default: False
# c.ServerApp.allow_remote_access = False

## Whether to allow the user to run the server as root.
#  Default: False
# c.ServerApp.allow_root = False

## Allow unauthenticated access to endpoints without authentication rule.
#  
#          When set to `True` (default in jupyter-server 2.0, subject to change
#          in the future), any request to an endpoint without an authentication rule
#          (either `@tornado.web.authenticated`, or `@allow_unauthenticated`)
#          will be permitted, regardless of whether user has logged in or not.
#  
#          When set to `False`, logging in will be required for access to each endpoint,
#          excluding the endpoints marked with `@allow_unauthenticated` decorator.
#  
#          This option can be configured using `JUPYTER_SERVER_ALLOW_UNAUTHENTICATED_ACCESS`
#          environment variable: any non-empty value other than "true" and "yes" will
#          prevent unauthenticated access to endpoints without `@allow_unauthenticated`.
#  Default: True
# c.ServerApp.allow_unauthenticated_access = True

## Answer yes to any prompts.
#  See also: JupyterApp.answer_yes
# c.ServerApp.answer_yes = False

## "
#          Require authentication to access prometheus metrics.
#  Default: True
# c.ServerApp.authenticate_prometheus = True

## The authorizer class to use.
#  Default: 'jupyter_server.auth.authorizer.AllowAllAuthorizer'
# c.ServerApp.authorizer_class = 'jupyter_server.auth.authorizer.AllowAllAuthorizer'

## Reload the webapp when changes are made to any Python src files.
#  Default: False
# c.ServerApp.autoreload = False

## The base URL for the Jupyter server.
#  
#                         Leading and trailing slashes can be omitted,
#                         and will automatically be added.
#  Default: '/'
# c.ServerApp.base_url = '/'

## Specify what command to use to invoke a web
#                        browser when starting the server. If not specified, the
#                        default browser will be determined by the `webbrowser`
#                        standard library module, which allows setting of the
#                        BROWSER environment variable to override it.
#  Default: ''
# c.ServerApp.browser = ''

## The full path to an SSL/TLS certificate file.
#  Default: ''
# c.ServerApp.certfile = ''

## The full path to a certificate authority certificate for SSL/TLS client
#  authentication.
#  Default: ''
# c.ServerApp.client_ca = ''

## Full path of a config file.
#  See also: JupyterApp.config_file
# c.ServerApp.config_file = ''

## Specify a config file to load.
#  See also: JupyterApp.config_file_name
# c.ServerApp.config_file_name = ''

## The config manager class to use
#  Default: 'jupyter_server.services.config.manager.ConfigManager'
# c.ServerApp.config_manager_class = 'jupyter_server.services.config.manager.ConfigManager'

## The content manager class to use.
#  Default: 'jupyter_server.services.contents.largefilemanager.AsyncLargeFileManager'
# c.ServerApp.contents_manager_class = 'jupyter_server.services.contents.largefilemanager.AsyncLargeFileManager'

## DEPRECATED. Use IdentityProvider.cookie_options
#  Default: {}
# c.ServerApp.cookie_options = {}

## The random bytes used to secure cookies.
#          By default this is generated on first start of the server and persisted across server
#          sessions by writing the cookie secret into the `cookie_secret_file` file.
#          When using an executable config file you can override this to be random at each server restart.
#  
#          Note: Cookie secrets should be kept private, do not share config files with
#          cookie_secret stored in plaintext (you can read the value from a file).
#  Default: b''
# c.ServerApp.cookie_secret = b''

## The file where the cookie secret is stored.
#  Default: ''
# c.ServerApp.cookie_secret_file = ''

## Override URL shown to users.
#  
#          Replace actual URL, including protocol, address, port and base URL,
#          with the given value when displaying URL to the users. Do not change
#          the actual connection URL. If authentication token is enabled, the
#          token is added to the custom URL automatically.
#  
#          This option is intended to be used when the URL to display to the user
#          cannot be determined reliably by the Jupyter server (proxified
#          or containerized setups for example).
#  Default: ''
# c.ServerApp.custom_display_url = ''

## The default URL to redirect to from `/`
#  Default: '/'
# c.ServerApp.default_url = '/'

## Disable cross-site-request-forgery protection
#  
#          Jupyter server includes protection from cross-site request forgeries,
#          requiring API requests to either:
#  
#          - originate from pages served by this server (validated with XSRF cookie and token), or
#          - authenticate with a token
#  
#          Some anonymous compute resources still desire the ability to run code,
#          completely without authentication.
#          These services can disable all authentication and security checks,
#          with the full knowledge of what that implies.
#  Default: False
# c.ServerApp.disable_check_xsrf = False

## The directory to look at for external kernel connection files, if
#  allow_external_kernels is True. Defaults to Jupyter
#  runtime_dir/external_kernels. Make sure that this directory is not filled with
#  left-over connection files, that could result in unnecessary kernel manager
#  creations.
#  Default: None
# c.ServerApp.external_connection_dir = None

## handlers that should be loaded at higher priority than the default services
#  Default: []
# c.ServerApp.extra_services = []

## Extra paths to search for serving static files.
#  
#          This allows adding javascript/css to be available from the Jupyter server machine,
#          or overriding individual files in the IPython
#  Default: []
# c.ServerApp.extra_static_paths = []

## Extra paths to search for serving jinja templates.
#  
#          Can be used to override templates from jupyter_server.templates.
#  Default: []
# c.ServerApp.extra_template_paths = []

## Open the named file when the application is launched.
#  Default: ''
# c.ServerApp.file_to_run = ''

## The URL prefix where files are opened directly.
#  Default: 'notebooks'
# c.ServerApp.file_url_prefix = 'notebooks'

## Generate default config file.
#  See also: JupyterApp.generate_config
# c.ServerApp.generate_config = False

## DEPRECATED. Use IdentityProvider.get_secure_cookie_kwargs
#  Default: {}
# c.ServerApp.get_secure_cookie_kwargs = {}

## The identity provider class to use.
#  Default: 'jupyter_server.auth.identity.PasswordIdentityProvider'
# c.ServerApp.identity_provider_class = 'jupyter_server.auth.identity.PasswordIdentityProvider'

## DEPRECATED. Use ZMQChannelsWebsocketConnection.iopub_data_rate_limit
#  Default: 0.0
# c.ServerApp.iopub_data_rate_limit = 0.0

## DEPRECATED. Use ZMQChannelsWebsocketConnection.iopub_msg_rate_limit
#  Default: 0.0
# c.ServerApp.iopub_msg_rate_limit = 0.0

## The IP address the Jupyter server will listen on.
#  Default: 'localhost'
# c.ServerApp.ip = 'localhost'

## Supply extra arguments that will be passed to Jinja environment.
#  Default: {}
# c.ServerApp.jinja_environment_options = {}

## Extra variables to supply to jinja templates when rendering.
#  Default: {}
# c.ServerApp.jinja_template_vars = {}

## Dict of Python modules to load as Jupyter server extensions.Entry values can
#  be used to enable and disable the loading ofthe extensions. The extensions
#  will be loaded in alphabetical order.
#  Default: {}
# c.ServerApp.jpserver_extensions = {}

## The kernel manager class to use.
#  Default: 'jupyter_server.services.kernels.kernelmanager.MappingKernelManager'
# c.ServerApp.kernel_manager_class = 'jupyter_server.services.kernels.kernelmanager.MappingKernelManager'

## The kernel spec manager class to use. Should be a subclass of
#  `jupyter_client.kernelspec.KernelSpecManager`.
#  
#  The Api of KernelSpecManager is provisional and might change without warning
#  between this version of Jupyter and the next stable one.
#  Default: 'builtins.object'
# c.ServerApp.kernel_spec_manager_class = 'builtins.object'

## The kernel websocket connection class to use.
#  Default: 'jupyter_server.services.kernels.connection.base.BaseKernelWebsocketConnection'
# c.ServerApp.kernel_websocket_connection_class = 'jupyter_server.services.kernels.connection.base.BaseKernelWebsocketConnection'

## DEPRECATED. Use ZMQChannelsWebsocketConnection.kernel_ws_protocol
#  Default: ''
# c.ServerApp.kernel_ws_protocol = ''

## The full path to a private key file for usage with SSL/TLS.
#  Default: ''
# c.ServerApp.keyfile = ''

## DEPRECATED. Use ZMQChannelsWebsocketConnection.limit_rate
#  Default: False
# c.ServerApp.limit_rate = False

## Hostnames to allow as local when allow_remote_access is False.
#  
#         Local IP addresses (such as 127.0.0.1 and ::1) are automatically accepted
#         as local as well.
#  Default: ['localhost']
# c.ServerApp.local_hostnames = ['localhost']

## The date format used by logging formatters for %(asctime)s
#  See also: Application.log_datefmt
# c.ServerApp.log_datefmt = '%Y-%m-%d %H:%M:%S'

## The Logging format template
#  See also: Application.log_format
# c.ServerApp.log_format = '[%(name)s]%(highlevel)s %(message)s'

## Set the log level by value or name.
#  See also: Application.log_level
# c.ServerApp.log_level = 30

## 
#  See also: Application.logging_config
# c.ServerApp.logging_config = {}

## The login handler class to use.
#  Default: 'jupyter_server.auth.login.LegacyLoginHandler'
# c.ServerApp.login_handler_class = 'jupyter_server.auth.login.LegacyLoginHandler'

## The logout handler class to use.
#  Default: 'jupyter_server.auth.logout.LogoutHandler'
# c.ServerApp.logout_handler_class = 'jupyter_server.auth.logout.LogoutHandler'

## Sets the maximum allowed size of the client request body, specified in the
#  Content-Length request header field. If the size in a request exceeds the
#  configured value, a malformed HTTP message is returned to the client.
#  
#  Note: max_body_size is applied even in streaming mode.
#  Default: 536870912
# c.ServerApp.max_body_size = 536870912

## Gets or sets the maximum amount of memory, in bytes, that is allocated for use
#  by the buffer manager.
#  Default: 536870912
# c.ServerApp.max_buffer_size = 536870912

## Gets or sets a lower bound on the open file handles process resource limit.
#  This may need to be increased if you run into an OSError: [Errno 24] Too many
#  open files. This is not applicable when running on Windows.
#  Default: 0
# c.ServerApp.min_open_files_limit = 0

## DEPRECATED, use root_dir.
#  Default: ''
# c.ServerApp.notebook_dir = ''

## Whether to open in a browser after starting.
#                          The specific browser used is platform dependent and
#                          determined by the python standard library `webbrowser`
#                          module, unless it is overridden using the --browser
#                          (ServerApp.browser) configuration option.
#  Default: False
# c.ServerApp.open_browser = False

## DEPRECATED in 2.0. Use PasswordIdentityProvider.hashed_password
#  Default: ''
# c.ServerApp.password = ''

## DEPRECATED in 2.0. Use PasswordIdentityProvider.password_required
#  Default: False
# c.ServerApp.password_required = False

## The port the server will listen on (env: JUPYTER_PORT).
#  Default: 0
# c.ServerApp.port = 0

## The number of additional ports to try if the specified port is not available
#  (env: JUPYTER_PORT_RETRIES).
#  Default: 50
# c.ServerApp.port_retries = 50

## Preferred starting directory to use for notebooks and kernels.
#  ServerApp.preferred_dir is deprecated in jupyter-server 2.0. Use
#  FileContentsManager.preferred_dir instead
#  Default: ''
# c.ServerApp.preferred_dir = ''

## DISABLED: use %pylab or %matplotlib in the notebook to enable matplotlib.
#  Default: 'disabled'
# c.ServerApp.pylab = 'disabled'

## If True, display controls to shut down the Jupyter server, such as menu items
#  or buttons.
#  Default: True
# c.ServerApp.quit_button = True

## DEPRECATED. Use ZMQChannelsWebsocketConnection.rate_limit_window
#  Default: 0.0
# c.ServerApp.rate_limit_window = 0.0

## Reraise exceptions encountered loading server extensions?
#  Default: False
# c.ServerApp.reraise_server_extension_failures = False

## The directory to use for notebooks and kernels.
#  Default: ''
# c.ServerApp.root_dir = ''

## The session manager class to use.
#  Default: 'builtins.object'
# c.ServerApp.session_manager_class = 'builtins.object'

## Instead of starting the Application, dump configuration to stdout
#  See also: Application.show_config
# c.ServerApp.show_config = False

## Instead of starting the Application, dump configuration to stdout (as JSON)
#  See also: Application.show_config_json
# c.ServerApp.show_config_json = False

## Shut down the server after N seconds with no kernelsrunning and no activity.
#  This can be used together with culling idle kernels
#  (MappingKernelManager.cull_idle_timeout) to shutdown the Jupyter server when
#  it's not in use. This is not precisely timed: it may shut down up to a minute
#  later. 0 (the default) disables this automatic shutdown.
#  Default: 0
# c.ServerApp.shutdown_no_activity_timeout = 0

## The UNIX socket the Jupyter server will listen on.
#  Default: ''
# c.ServerApp.sock = ''

## The permissions mode for UNIX socket creation (default: 0600).
#  Default: '0600'
# c.ServerApp.sock_mode = '0600'

## Supply SSL options for the tornado HTTPServer.
#              See the tornado docs for details.
#  Default: {}
# c.ServerApp.ssl_options = {}

## Paths to set up static files as immutable.
#  
#  This allow setting up the cache control of static files as immutable. It
#  should be used for static file named with a hash for instance.
#  Default: []
# c.ServerApp.static_immutable_cache = []

## Supply overrides for terminado. Currently only supports "shell_command".
#  Default: {}
# c.ServerApp.terminado_settings = {}

## Set to False to disable terminals.
#  
#           This does *not* make the server more secure by itself.
#           Anything the user can in a terminal, they can also do in a notebook.
#  
#           Terminals may also be automatically disabled if the terminado package
#           is not available.
#  Default: False
# c.ServerApp.terminals_enabled = False

## DEPRECATED. Use IdentityProvider.token
#  Default: '<DEPRECATED>'
# c.ServerApp.token = '<DEPRECATED>'
# !!!!! ---- Disable token/password authentication ------ !!!!
c.ServerApp.password = ''
c.ServerApp.token = ''

## Supply overrides for the tornado.web.Application that the Jupyter server uses.
#  Default: {}
# c.ServerApp.tornado_settings = {}

## Whether to trust or not X-Scheme/X-Forwarded-Proto and X-Real-Ip/X-Forwarded-
#  For headerssent by the upstream reverse proxy. Necessary if the proxy handles
#  SSL
#  Default: False
# c.ServerApp.trust_xheaders = False

## Disable launching browser by redirect file
#       For versions of notebook > 5.7.2, a security feature measure was added that
#       prevented the authentication token used to launch the browser from being visible.
#       This feature makes it difficult for other users on a multi-user system from
#       running code in your Jupyter session as you.
#       However, some environments (like Windows Subsystem for Linux (WSL) and Chromebooks),
#       launching a browser using a redirect file can lead the browser failing to load.
#       This is because of the difference in file structures/paths between the runtime and
#       the browser.
#  
#       Disabling this setting to False will disable this behavior, allowing the browser
#       to launch by using a URL and visible token (as before).
#  Default: True
# c.ServerApp.use_redirect_file = True

## Specify where to open the server on startup. This is the
#          `new` argument passed to the standard library method `webbrowser.open`.
#          The behaviour is not guaranteed, but depends on browser support. Valid
#          values are:
#  
#           - 2 opens a new tab,
#           - 1 opens a new window,
#           - 0 opens in an existing window.
#  
#          See the `webbrowser.open` documentation for details.
#  Default: 2
# c.ServerApp.webbrowser_open_new = 2

## Set the tornado compression options for websocket connections.
#  
#  This value will be returned from
#  :meth:`WebSocketHandler.get_compression_options`. None (default) will disable
#  compression. A dict (even an empty one) will enable compression.
#  
#  See the tornado docs for WebSocketHandler.get_compression_options for details.
#  Default: None
# c.ServerApp.websocket_compression_options = None

## Configure the websocket ping interval in seconds.
#  
#  Websockets are long-lived connections that are used by some Jupyter Server
#  extensions.
#  
#  Periodic pings help to detect disconnected clients and keep the connection
#  active. If this is set to None, then no pings will be performed.
#  
#  When a ping is sent, the client has ``websocket_ping_timeout`` seconds to
#  respond. If no response is received within this period, the connection will be
#  closed from the server side.
#  Default: 0
# c.ServerApp.websocket_ping_interval = 0

## Configure the websocket ping timeout in seconds.
#  
#  See ``websocket_ping_interval`` for details.
#  Default: 0
# c.ServerApp.websocket_ping_timeout = 0

## The base URL for websockets,
#          if it differs from the HTTP server (hint: it almost certainly doesn't).
#  
#          Should be in the form of an HTTP origin: ws[s]://hostname[:port]
#  Default: ''
# c.ServerApp.websocket_url = ''



from dataclasses import dataclass


@dataclass
class Constants:
    # declare const user config variables for email notification
    is_serverless = "spark.expectations.is.serverless"
    se_notifications_enable_smtp_server_auth = "spark.expectations.notifications.email.smtp.server.auth"
    se_notifications_smtp_password = "spark.expectations.notifications.email.smtp.password"
    se_notifications_smtp_creds_dict = "spark.expectations.notifications.smtp.creds.dict"
    cbs_smtp_password = "spark.expectations.notifications.cerberus.smtp.password"
    dbx_smtp_password = "spark.expectations.notifications.dbx.smtp.password"

    se_user_defined_custom_dataframe = "spark.expectations.user.defined.custom.dataframe"
    se_notifications_enable_custom_dataframe = "spark.expectations.notifications.enable.custom.dataframe"
    se_dq_obs_default_email_template = "spark.expectations.dq.obs.default.detailed.email.template"
    se_notifications_default_basic_email_template = "spark.expectations.notifications.email.default.basic.template"
    se_dq_obs_mode_of_communication = "spark.expectations.dq.obs.mode.of.communication"
    se_notifications_service_account_email = "spark.expectations.notifications.service.account.email"
    se_dq_obs_alert_flag = "spark.expectations.notifications.alert.flag.disable"
    se_notifications_service_account_password = "spark.expectations.notifications.service.account.password"
    se_notifications_smtp_user_name = "spark.expectations.notifications.smtp.user.name"
    se_notifications_enable_email = "spark.expectations.notifications.email.enabled"
    se_enable_obs_dq_report_result = "spark.expectations.notifications.observability.enabled"
    se_notifications_enable_custom_email_body = "spark.expectations.notifications.email.custom.body.enable"
    se_notifications_email_smtp_host = "spark.expectations.notifications.email.smtp.host"
    se_notifications_email_smtp_port = "spark.expectations.notifications.email.smtp.port"
    se_notifications_email_from = "spark.expectations.notifications.email.from"
    se_notifications_email_to_other_mail_id = "spark.expectations.notifications.email.to.other.mail.com"
    se_notifications_email_subject = "spark.expectations.notifications.email.subject"
    se_notifications_email_custom_body = "spark.expectations.notifications.email.custom.body"
    se_notifications_enable_templated_basic_email_body = (
        "spark.expectations.notifications.email.templated.basic.body.enable"
    )
    se_notifications_enable_templated_custom_email = (
        "spark.expectations.notifications.email.templated.custom.email.enable"
    )
    se_notifications_email_custom_template = "spark.expectations.notifications.email.custom.template"

    # declare const user config variables for slack notification
    se_notifications_enable_slack = "spark.expectations.notifications.slack.enabled"
    se_notifications_slack_webhook_url = "spark.expectations.notifications.slack.webhook.url"

    # declare const user config variables for teams notification
    se_notifications_enable_teams = "spark.expectations.notifications.teams.enabled"
    se_notifications_teams_webhook_url = "spark.expectations.notifications.teams.webhook.url"

    # declare const user config variables for zoom notification
    se_notifications_enable_zoom = "spark.expectations.notifications.zoom.enabled"
    se_notifications_zoom_webhook_url = "spark.expectations.notifications.zoom.webhook.url"
    se_notifications_zoom_token = "spark.expectations.notifications.zoom.token"

    # declare const user configs for pagerduty incidents
    se_notifications_enable_pagerduty = "spark.expectations.notifications.pagerduty.enabled"
    # For Pagerduty Services, your integration key is the routing key when making calls to the Events API v2.
    se_notifications_pagerduty_integration_key = "spark.expectations.notifications.pagerduty.integration.key"
    se_notifications_pagerduty_webhook_url = "spark.expectations.notifications.pagerduty.webhook.url"

    se_notifications_on_start = "spark.expectations.notifications.on.start"
    se_notifications_on_completion = "spark.expectations.notifications.on.completion"
    se_notifications_on_fail = "spark.expectations.notifications.on.fail"
    se_notifications_on_error_drop_exceeds_threshold_breach = (
        "spark.expectations.notifications.on.error.drop.exceeds.threshold.breach"
    )
    se_notifications_on_rules_action_if_failed_set_ignore = (
        "spark.expectations.notifications.on.rules.action.if.failed.set.ignore"
    )
    se_notifications_on_error_drop_threshold = "spark.expectations.notifications.error.drop.threshold"
    se_notifications_min_priority_slack = "spark.expectations.notifications.slack.min.priority"

    se_enable_streaming = "se.streaming.enable"
    se_enable_error_table = "se.enable.error.table"
    se_dq_rules_params = "se.dq.rules.params"
    se_streaming_stats_kafka_custom_config_enable = "se.streaming.stats.kafka.custom.config.enable"
    se_streaming_stats_topic_name = "se.streaming.stats.topic.name"
    se_streaming_stats_kafka_bootstrap_server = "se.streaming.stats.kafka.bootstrap.server"

    secret_type = "se.streaming.secret.type"

    cbs_url = "se.streaming.cerberus.url"
    cbs_sdb_path = "se.streaming.cerberus.sdb.path"
    cbs_kafka_server_url = "se.streaming.cerberus.kafka.server.url"
    cbs_secret_token_url = "se.streaming.cbs.secret.token.url"
    cbs_secret_app_name = "se.streaming.cbs.secret.app.name"
    cbs_secret_token = "se.streaming.cerberus.secret.token"
    cbs_topic_name = "se.streaming.cerberus.token.name"

    dbx_workspace_url = "se.streaming.dbx.workspace.url"
    dbx_secret_scope = "se.streaming.dbx.secret.scope"
    dbx_kafka_server_url = "se.streaming.dbx.kafka.server.url"
    dbx_secret_token_url = "se.streaming.dbx.secret.token.url"
    dbx_secret_app_name = "se.streaming.dbx.secret.app.name"
    dbx_secret_token = "se.streaming.dbx.secret.token"
    dbx_topic_name = "se.streaming.dbx.topic.name"

    # declare const user config variables for agg query dq detailed stats
    se_enable_agg_dq_detailed_result = "spark.expectations.agg.dq.detailed.stats"
    se_enable_query_dq_detailed_result = "spark.expectations.query.dq.detailed.stats"
    se_job_metadata = "spark.expectations.job.metadata"

    querydq_output_custom_table_name = "spark.expectations.query.dq.custom.table_name"

    # declare const variable for agg query dq detailed stats

    se_agg_dq_expectation_regex_pattern = r"(\(.+?\)|\w+\(.+?\))(\s*[<>!=]+\s*.+|\s*between\s*.+)$"
    # declare const variable for range in agg query dq detailed stats
    # ex). count(*), count(), count(col1), sum(col1), col1
    allowed_functions = r"(\w+\(\*\)|\w+\(\w+\)|\w+)"
    and_clause = r"(\s+and\s+)"
    # ex). col1 > 1, col 1 < 10
    operator_with_value = r"(\s*[><]\s*\d+)"

    se_agg_dq_expectation_range_regex_pattern = (
        allowed_functions + operator_with_value + and_clause + allowed_functions + operator_with_value
    )


from spark_expectations.sinks import get_sink_hook
from spark_expectations.sinks.plugins.kafka_writer import (
    SparkExpectationsKafkaWritePluginImpl,
)

def test_get_sink_hook():
    pm = get_sink_hook()
    # Check that the correct number of plugins have been registered
    assert len(pm.list_name_plugin()) == 1

    # Check that the correct plugins have been registered
    assert isinstance(pm.get_plugin("spark_expectations_kafka_write"), SparkExpectationsKafkaWritePluginImpl)



import pluggy
from spark_expectations.notifications import get_notifications_hook
from spark_expectations.notifications.plugins.email import (
    SparkExpectationsEmailPluginImpl,
)
from spark_expectations.notifications.plugins.slack import (
    SparkExpectationsSlackPluginImpl,
)
from spark_expectations.notifications.plugins.teams import (
    SparkExpectationsTeamsPluginImpl,
)
from spark_expectations.notifications.plugins.zoom import (
    SparkExpectationsZoomPluginImpl,
)
from spark_expectations.notifications.plugins.pagerduty import (
    SparkExpectationsPagerDutyPluginImpl,
)


def test_notifications_hook():
    pm = get_notifications_hook()
    pm.get_plugins()

    # act
    email_plugin = pm.get_plugin("spark_expectations_email_notification")
    slack_plugin = pm.get_plugin("spark_expectations_slack_notification")
    teams_plugin = pm.get_plugin("spark_expectations_teams_notification")
    zoom_plugin = pm.get_plugin("spark_expectations_zoom_notification")
    pagerduty_plugin = pm.get_plugin("spark_expectations_pagerduty_notification")
    # Check that the correct number of plugins have been registered
    assert len(pm.list_name_plugin()) == 5
    # assert
    assert isinstance(pm, pluggy.PluginManager)
    assert email_plugin is not None
    assert slack_plugin is not None
    assert teams_plugin is not None
    assert zoom_plugin is not None
    assert pagerduty_plugin is not None
    assert isinstance(email_plugin, SparkExpectationsEmailPluginImpl)
    assert isinstance(slack_plugin, SparkExpectationsSlackPluginImpl)
    assert isinstance(teams_plugin, SparkExpectationsTeamsPluginImpl)
    assert isinstance(zoom_plugin, SparkExpectationsZoomPluginImpl)
    assert isinstance(pagerduty_plugin, SparkExpectationsPagerDutyPluginImpl)



from typing import Dict, Union
from pyspark.sql import DataFrame
from spark_expectations.sinks.plugins.base_writer import (
    SparkExpectationsSinkWriter,
    spark_expectations_writer_impl,
)
from spark_expectations.core.exceptions import SparkExpectationsMiscException
from spark_expectations import _log


class SparkExpectationsKafkaWritePluginImpl(SparkExpectationsSinkWriter):
    """
    class helps to write the stats data into the NSP
    """

    @spark_expectations_writer_impl
    def writer(self, _write_args: Dict[Union[str], Union[str, bool, Dict[str, str], DataFrame]]) -> None:
        """
        The functions helps to write data into the kafka topic
        Args:
            _write_args:

        Returns:

        """

        try:
            # kafka_options = {
            #     "kafka.bootstrap.servers": "localhost:9092",
            #     "topic": _write_args.get("nsp_topic_name"),
            #     "failOnDataLoss": "true",
            # }

            _log.debug(f"_write_args: {_write_args}")

            if _write_args.pop("enable_se_streaming"):
                _log.info("started write stats data into kafka stats topic")

                df: DataFrame = _write_args.get("stats_df")
                kafka_options = _write_args.get("kafka_write_options")
                
                # Log Kafka connection details (mask sensitive info)
                masked_options = {k: (v if "password" not in k.lower() and "secret" not in k.lower() and "token" not in k.lower() else "***MASKED***") for k, v in kafka_options.items()}
                _log.info(f"Writing to Kafka with options: {masked_options}")

                df.selectExpr("to_json(struct(*)) AS value").write.format("kafka").mode("append").options(
                    **kafka_options
                ).save()

                _log.info("ended writing stats data into kafka stats topic")

        except Exception as e:
            # Log detailed error information
            kafka_options = _write_args.get("kafka_write_options", {})
            topic = kafka_options.get("topic", "unknown")
            bootstrap_servers = kafka_options.get("kafka.bootstrap.servers", "unknown")
            
            error_details = f"Failed to write to Kafka topic '{topic}' on servers '{bootstrap_servers}'. Error: {str(e)}"
            _log.error(f"Kafka write failure: {error_details}")
            
            raise SparkExpectationsMiscException(f"error occurred while saving data into kafka: {error_details}")
